# test_configparser_subprocess.py
#
# -*- coding: utf-8 -*-
#
#  Copyright (C) 2010-2012, 2015, 2018, 2019  University of Zurich. All rights reserved.
#
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or (at your
#  option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

# stdlib imports
from __future__ import absolute_import, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import range
from builtins import object
from builtins import str

from io import StringIO
import os
import shutil
import tempfile
import re
from itertools import product

import mock
import pytest

# GC3Pie imports
from gc3libs import Run, Application
import gc3libs.config
import gc3libs.core
import gc3libs.template
from gc3libs.quantity import GB, hours
from gc3libs.backends.shellcmd import ShellcmdLrms
from gc3libs.backends.transport import SshTransport
from gc3libs.quantity import Memory, Duration

from gc3libs.testing.helpers import temporary_directory


def _setup_config_file(confstr):
    (fd, name) = tempfile.mkstemp()
    f = os.fdopen(fd, 'w+')
    f.write(confstr)
    f.close()
    return name


def import_invalid_conf(confstr, **extra_args):
    """`load` just ignores invalid input."""
    tmpfile = _setup_config_file(confstr)
    cfg = gc3libs.config.Configuration()
    try:
        cfg.load(tmpfile)
    finally:
        os.remove(tmpfile)
    assert len(cfg.resources) == 0
    assert len(cfg.auths) == 1
    assert 'none' in cfg.auths


# pylint: disable=unused-argument
def read_invalid_conf(confstr, **extra_args):
    """`merge_file` raises a `ConfigurationError` exception on invalid input"""
    tmpfile = _setup_config_file(confstr)
    cfg = gc3libs.config.Configuration()
    try:
        cfg.merge_file(tmpfile)
    finally:
        os.remove(tmpfile)


# pylint: disable=unused-argument
def parse_invalid_conf(confstr, **extra_args):
    """`_parse` raises a `ConfigurationError` exception on invalid input."""
    cfg = gc3libs.config.Configuration()
    # pylint: disable=unused-variable,protected-access
    defaults, resources, auths = cfg._parse(StringIO(confstr))


def test_valid_conf():
    """Test parsing a valid configuration file"""
    tmpfile = _setup_config_file("""
[auth/ssh]
type = ssh
username = gc3pie

[resource/test]
type = shellcmd
auth = ssh
transport = local
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8
max_cores = 2
architecture = x86_64
override = False
    """)
    try:
        cfg = gc3libs.config.Configuration(tmpfile)
        resources = cfg.make_resources(ignore_errors=False)
        # resources are enabled by default
        assert 'test' in resources
        assert isinstance(resources['test'], ShellcmdLrms)
        # test types
        assert isinstance(resources['test']['name'], str)
        assert isinstance(resources['test']['max_cores_per_job'], int)
        assert isinstance(resources['test']['max_memory_per_core'], Memory)
        assert isinstance(resources['test']['max_walltime'], Duration)
        assert isinstance(resources['test']['max_cores'], int)
        assert isinstance(resources['test']['architecture'], set)
        # test parsed values
        assert resources['test']['name'] == 'test'
        assert resources['test']['max_cores_per_job'] == 2
        assert resources['test']['max_memory_per_core'] == 2 * GB
        assert resources['test']['max_walltime'] == 8 * hours
        assert resources['test']['max_cores'] == 2
        assert (resources['test']['architecture'] ==
                     set([Run.Arch.X86_64]))
    finally:
        os.remove(tmpfile)


invalid_confs = [
    # #0
    (
        'no section header',
        "THIS IS NOT A CONFIG FILE"
    ),
    # #1
    (
    'malformed key=value',
    """
[auth/ssh]
type - ssh
            """
    ),
    # #2
    (
        'missing required key `type` in `auth` section',
        """
[auth/ssh]
#type = ssh
        """
    ),
    # #3
    (
        'missing required key `type` in `resource` section',
        """
[resource/test]
#type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 2GiB
max_walltime = 8 hours
max_cores = 10
architecture = x86_64
        """
    ),
    (
        'missing required key `max_cores_per_job` in `resource` section',
        """
[resource/test]
type = shellcmd
frontend = localhost
transport = local
#max_cores_per_job = 2
max_memory_per_core = 2GiB
max_walltime = 8 hours
max_cores = 10
architecture = x86_64
        """
    ),
    (
        'missing required key `max_memory_per_core` in `resource` section',
        """
[resource/test]
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 2
#max_memory_per_core = 2GiB
max_walltime = 8 hours
max_cores = 10
architecture = x86_64
        """
    ),
    (
        'missing required key `max_walltime` in `resource` section',
        """
[resource/test]
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 2GiB
#max_walltime = 8 hours
max_cores = 10
architecture = x86_64
        """
    ),
    (
        'missing required key `max_cores` in `resource` section',
        """
[resource/test]
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 2GiB
max_walltime = 8 hours
#max_cores = 10
architecture = x86_64
        """
    ),
    (
        'missing required key `architecture` in `resource` section',
        """
[resource/test]
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 2GiB
max_walltime = 8 hours
max_cores = 10
#architecture = x86_64
        """
    )
]

@pytest.mark.parametrize("conf", invalid_confs)
def test_import_invalid_confs(conf):
    """Test reading invalid configuration files"""
    with pytest.raises(gc3libs.exceptions.NoConfigurationFile):
        import_invalid_conf(conf[1])

@pytest.mark.parametrize("conf", invalid_confs)
def test_read_invalid_confs(conf):
    """Test reading invalid configuration files"""
    with pytest.raises(gc3libs.exceptions.ConfigurationError):
        read_invalid_conf(conf[1])

@pytest.mark.parametrize("conf", invalid_confs)
def test_parse_invalid_confs(conf):
    """Test reading invalid configuration files"""
    with pytest.raises(gc3libs.exceptions.ConfigurationError):
        parse_invalid_conf(conf[1])



def test_load_non_existing_file():
    """Test that `Configuration.load()` raises a `NoAccessibleConfigurationFile` exception if no configuration file exists"""
    with pytest.raises(gc3libs.exceptions.NoAccessibleConfigurationFile):
        # pylint: disable=unused-variable
        cfg = gc3libs.config.Configuration('/NON_EXISTING_FILE')


def test_load_non_readable_file():
    """Test that `Configuration.load()` raises a `NoAccessibleConfigurationFile` exception if no configuration file can be read"""
    with tempfile.NamedTemporaryFile(prefix=__name__, mode='wt') as tmpfile:
        os.chmod(tmpfile.name, 0)
        with pytest.raises(gc3libs.exceptions.NoAccessibleConfigurationFile):
            # pylint: disable=unused-variable
            cfg = gc3libs.config.Configuration(tmpfile.name)


def test_load_non_valid_file():
    """Test that `Configuration.load()` raises a `NoValidConfigurationFile` exception if no configuration file can be parsed"""
    tmpfile = _setup_config_file("""
This is not a valid configuration file.
""")
    try:
        with pytest.raises(gc3libs.exceptions.NoValidConfigurationFile):
            # pylint: disable=unused-variable
            cfg = gc3libs.config.Configuration(tmpfile)
    finally:
        os.remove(tmpfile)


def test_auth_none():
    """Test that `auth = none` is always available."""
    tmpfile = _setup_config_file("""
[resource/test]
type = pbs
auth = none
frontend = localhost
transport = local
max_cores_per_job = 44
max_memory_per_core = 55 GB
max_walltime = 66 hours
max_cores = 77
architecture = x86_64
    """)
    try:
        cfg = gc3libs.config.Configuration(tmpfile)
        auth = cfg.make_auth('none')
        assert isinstance(auth(), gc3libs.authentication.NoneAuth)
        # if this doesn't raise any error, we're good
        resources = cfg.make_resources(ignore_errors=False)
    finally:
        os.remove(tmpfile)


def test_auth_empty_keypair_name():
    """Test that if keypair_name is specified and empty, we fail."""
    tmpfile = _setup_config_file("""
[resource/kptest]
type = shellcmd
auth = ssh
keypair_name =
    """)
    try:
        with pytest.raises(gc3libs.exceptions.NoValidConfigurationFile):
            # pylint: disable=unused-variable
            cfg = gc3libs.config.Configuration(tmpfile)
    finally:
        os.remove(tmpfile)


def test_key_renames():
    """Test that `ncores` is renamed to `max_cores` during parse"""
    tmpfile = _setup_config_file("""
[auth/ssh]
type = ssh
username = gc3pie

[resource/test]
type = shellcmd
auth = ssh
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8
# `ncores` renamed to `max_cores`
ncores = 77
architecture = x86_64
override = False
    """)
    try:
        cfg = gc3libs.config.Configuration(tmpfile)
        resources = cfg.make_resources(ignore_errors=False)
        assert 'ncores' not in resources['test']
        assert 'max_cores' in resources['test']
        assert resources['test']['max_cores'] == 77
    finally:
        os.remove(tmpfile)


class TestReadMultiple(object):

    @pytest.fixture(autouse=True)
    def setUp(self):
        self.f1 = _setup_config_file("""
[resource/localhost]
seq = 1
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 2GiB
max_walltime = 8 hours
max_cores = 10
architecture = x86_64
        """)
        self.f2 = _setup_config_file("""
[resource/localhost]
seq = 1
foo = 2
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 2GiB
max_walltime = 8 hours
max_cores = 10
architecture = x86_64
        """)
        self.f3 = _setup_config_file("""
[resource/localhost]
seq = 3
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 2GiB
max_walltime = 8 hours
max_cores = 10
architecture = x86_64
        """)

        yield
        # Finalize
        os.remove(self.f1)
        os.remove(self.f2)
        os.remove(self.f3)

    def test_read_config_multiple(self):
        """Test that `Configuration` aggregates multiple files"""
        cfg = gc3libs.config.Configuration(self.f1, self.f2, self.f3)
        assert cfg.resources['localhost']['seq'] == '3'
        assert cfg.resources['localhost']['foo'] == '2'

    def test_load_multiple(self):
        """Test that `Configuration.load` aggregates multiple files"""
        cfg = gc3libs.config.Configuration()
        cfg.load(self.f1, self.f2, self.f3)
        assert cfg.resources['localhost']['seq'] == '3'
        assert cfg.resources['localhost']['foo'] == '2'

    def test_merge_file_multiple(self):
        """Test that `Configuration.load` aggregates multiple files"""
        cfg = gc3libs.config.Configuration()
        cfg.merge_file(self.f1)
        cfg.merge_file(self.f2)
        cfg.merge_file(self.f3)
        assert cfg.resources['localhost']['seq'] == '3'
        assert cfg.resources['localhost']['foo'] == '2'


def test_no_valid_config1():
    """`Configuration.load` raises an exception if called with no arguments"""
    cfg = gc3libs.config.Configuration()
    with pytest.raises(gc3libs.exceptions.NoConfigurationFile):
        cfg.load()


def test_no_valid_config2():
    """Test that `Configuration.load` raises an exception if none of the
    arguments is a valid config file"""
    f1 = _setup_config_file("INVALID INPUT")
    f2 = _setup_config_file("INVALID INPUT")
    cfg = gc3libs.config.Configuration()
    try:
        with pytest.raises(gc3libs.exceptions.NoConfigurationFile):
            cfg.load(f1, f2)
    finally:
        os.remove(f1)
        os.remove(f2)


valid_architectures = [
    # a sample of the architecture strings that we accept
    ('x86_64', [gc3libs.Run.Arch.X86_64]),
    ('x86 64-bit', [gc3libs.Run.Arch.X86_64]),
    ('64bits x86', [gc3libs.Run.Arch.X86_64]),
    ('amd64', [gc3libs.Run.Arch.X86_64]),
    ('AMD64', [gc3libs.Run.Arch.X86_64]),
    ('intel 64', [gc3libs.Run.Arch.X86_64]),
    ('emt64', [gc3libs.Run.Arch.X86_64]),
    ('64 bits', [gc3libs.Run.Arch.X86_64]),
    ('64-BIT', [gc3libs.Run.Arch.X86_64]),

    ('x86 32-bit', [gc3libs.Run.Arch.X86_32]),
    ('32bits x86', [gc3libs.Run.Arch.X86_32]),
    ('i386', [gc3libs.Run.Arch.X86_32]),
    ('i486', [gc3libs.Run.Arch.X86_32]),
    ('i586', [gc3libs.Run.Arch.X86_32]),
    ('i686', [gc3libs.Run.Arch.X86_32]),
    ('32 bits', [gc3libs.Run.Arch.X86_32]),
    ('32-bit', [gc3libs.Run.Arch.X86_32]),

    ('32bit, 64bit', [gc3libs.Run.Arch.X86_64, gc3libs.Run.Arch.X86_32]),
    ('64bit, x86_64', [gc3libs.Run.Arch.X86_64]),
]

@pytest.mark.parametrize("arch", valid_architectures)
def test_valid_architectures(arch):
    """Test that valid architecture strings are parsed correctly"""
    config_template = gc3libs.template.Template("""
[resource/test]
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 2GiB
max_walltime = 8 hours
max_cores = 10
architecture = ${arch}
""")
    _check_parse_arch(config_template.substitute(arch=arch[0]), arch[1])


def _check_parse_arch(confstr, result):
    cfg = gc3libs.config.Configuration()
    defaults, resources, auths = cfg._parse(StringIO(confstr))
    assert isinstance(resources['test']['architecture'], set)
    assert resources['test']['architecture'] == set(result)


invalid_architectures = [
    '96bits', '31-BITS', 'amd65', 'xyzzy'
]

@pytest.mark.parametrize("arch", invalid_architectures)
def test_invalid_architectures(arch):
    """Invalid arch. strings should be rejected with `ConfigurationError` """
    config = gc3libs.template.Template("""
[resource/test]
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 2GiB
max_walltime = 8 hours
max_cores = 10
architecture = ${arch}
""").substitute(arch=arch)
    with pytest.raises(gc3libs.exceptions.ConfigurationError):
        _check_parse_arch(config, "should not be used")


class TestPrologueEpilogueScripts(object):

    """
    Test `prologue` and `epilogue` options for batch backends
    """

    @pytest.fixture(autouse=True)
    def setUp(self):
        # setup conf dir and conf file
        self.files_to_remove = []

        self.conftmpdir = tempfile.mkdtemp()
        self.tmpdir = tempfile.mkdtemp()
        # setup config file and sctipts
        cfgfname = os.path.join(self.conftmpdir, 'gc3pie.conf')
        self.files_to_remove.append(self.conftmpdir)
        self.files_to_remove.append(self.tmpdir)

        cfgstring = """
[auth/ssh]
type = ssh
username = gc3pie

[resource/test]
type = shellcmd
auth = ssh
transport = local
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8
max_cores = 2
architecture = x86_64
prologue = %(tmpdir)s/scripts/shellcmd_pre.sh
epilogue = %(tmpdir)s/scripts/shellcmd_post.sh
myapp_prologue = %(tmpdir)s/scripts/myapp_shellcmd_pre.sh
myapp_epilogue = %(tmpdir)s/scripts/myapp_shellcmd_post.sh
override = False

[resource/testpbs]
type = pbs
auth = ssh
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8
max_cores = 2
architecture = x86_64
prologue = %(tmpdir)s/scripts/shellcmd_pre.sh
epilogue = %(tmpdir)s/scripts/shellcmd_post.sh
myapp_prologue = %(tmpdir)s/scripts/myapp_shellcmd_pre.sh
myapp_epilogue = %(tmpdir)s/scripts/myapp_shellcmd_post.sh
app2_prologue_content = echo prologue app2
app2_epilogue_content = echo epilogue app2
""" % {'tmpdir': self.tmpdir}

        fdcfg = open(cfgfname, 'w')
        fdcfg.write(cfgstring)
        fdcfg.close()
        self.cfg = gc3libs.config.Configuration(
            cfgfname,
            auto_enable_auth=True)

        self.scripts = [
            'prologue',
            'epilogue',
            'myapp_prologue',
            'myapp_epilogue']
        os.mkdir(os.path.join(self.tmpdir, 'scripts'))
        for k, v in self.cfg['resources']['test'].items():
            if k in self.scripts:
                scriptfd = open(os.path.join(self.tmpdir, v), 'w')
                scriptfd.write('echo %s' % k)
                scriptfd.close()

        self.cfg = gc3libs.config.Configuration(cfgfname)
        # self.resources = self.cfg.make_resources(ignore_errors=False)
        # assert_equal(resources['test']['prologue'],
        #              'scripts/shellcmd_pre.sh')

        yield
        # Finalize
        for dirname in self.files_to_remove:
            shutil.rmtree(dirname)

    def test_scriptfiles_are_abs(self):
        """Test that prologue and epilogue scripts are absolute pathnames"""
        self.core = gc3libs.core.Core(self.cfg)
        for resource in self.core.get_resources():
            for (k, v) in resource.items():
                if k not in ['prologue', 'epilogue',
                             'myapp_prologue', 'myapp_epilogue']:
                    continue
                assert os.path.isfile(v)
                assert os.path.isabs(v)
                assert (os.path.abspath(v) ==
                             v)

    def test_pbs_prologue_and_epilogue_contents_when_files(self):
        """Prologue and epilogue scripts are inserted in the submission script
        """
        app = Application(['/bin/true'], [], [], '')
        self.core = gc3libs.core.Core(self.cfg)
        self.core.select_resource('testpbs')

        # get the selected resource
        lrms = [ resource for resource in self.core.get_resources() if resource.name == 'testpbs'][0]
        # Ugly hack. We have to list the job dirs to check which one
        # is the new one.
        jobdir = os.path.expandvars(lrms.spooldir)
        jobs = []
        if os.path.isdir(jobdir):
            jobs = os.listdir(jobdir)

        try:
            self.core.submit(app)
        except Exception:
            # it is normal to have an error since we don't probably
            # run a pbs server in this machine.
            pass

        newjobs = [d for d in os.listdir(jobdir) if d not in jobs]

        # There must be only one more job...
        assert len(newjobs) == 1

        newjobdir = os.path.join(jobdir, newjobs[0])
        self.files_to_remove.append(newjobdir)

        # and only one file in it
        assert len(os.listdir(newjobdir)) == 1

        # Check the content of the script file
        scriptfname = os.path.join(newjobdir, (os.listdir(newjobdir)[0]))
        scriptfile = open(scriptfname)
        assert re.match(
            "#!/bin/sh.*# prologue file `.*` BEGIN.*echo prologue.*# prologue"
            " file END.*/bin/true.*# epilogue file `.*` BEGIN.*echo epilogue"
            ".*# epilogue file END",
            scriptfile.read(),
            re.DOTALL | re.M)
        scriptfile.close()

        # kill the job
        if app.execution.state != Run.State.NEW:
            self.core.kill(app)

    def test_pbs_prologue_and_epilogue_contents_when_not_files(self):
        """Prologue and epilogue scripts are inserted in the submission script
        """
        app = Application(['/bin/true'], [], [], '')
        app.application_name = 'app2'
        self.core = gc3libs.core.Core(self.cfg)
        self.core.select_resource('testpbs')

        # get the selected resource
        lrms = [ resource for resource in self.core.get_resources() if resource.name == 'testpbs'][0]
        # Ugly hack. We have to list the job dirs to check which one
        # is the new one.
        jobdir = os.path.expandvars(lrms.spooldir)
        jobs = []
        if os.path.isdir(jobdir):
            jobs = os.listdir(jobdir)

        try:
            self.core.submit(app)
        except Exception:
            # it is normal to have an error since we don't probably
            # run a pbs server in this machine.
            pass

        newjobs = [d for d in os.listdir(jobdir) if d not in jobs]

        # There must be only one more job...
        assert len(newjobs) == 1

        newjobdir = os.path.join(jobdir, newjobs[0])
        self.files_to_remove.append(newjobdir)

        # and only one file in it
        assert len(os.listdir(newjobdir)) == 1

        # Check the content of the script file
        scriptfname = os.path.join(newjobdir, (os.listdir(newjobdir)[0]))
        scriptfile = open(scriptfname)
        assert re.match(
            "#!/bin/sh.*"
            "# prologue file `.*` BEGIN.*"
            "echo prologue.*"
            "# prologue file END.*"
            "# inline script BEGIN.*"
            "echo prologue app2.*"
            "# inline script END.*"
            "/bin/true.*"
            "# epilogue file `.*` BEGIN.*"
            "echo epilogue.*"
            "# epilogue file END.*"
            "# inline script BEGIN.*"
            "echo epilogue app2.*"
            "# inline script END",
            scriptfile.read(),
            re.DOTALL | re.M)
        scriptfile.close()

        # kill the job
        if app.execution.state != Run.State.NEW:
            self.core.kill(app)

    def test_prologue_epilogue_issue_352(self):
        cfgstring = """
[auth/ssh]
type = ssh
username = gc3pie

[resource/%s]
type = shellcmd
auth = ssh
transport = local
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8
max_cores = 2
architecture = x86_64
override = False
"""
        dir1, dir2 = tempfile.mkdtemp(), tempfile.mkdtemp()
        self.files_to_remove.extend([dir1, dir2])
        cfgfilenames = []

        for d in [dir1, dir2]:
            cfgfname = os.path.join(d, 'gc3pie.conf')
            fdcfg = open(cfgfname, 'w')
            fdcfg.write(cfgstring % os.path.basename(d))
            fdcfg.close()
            cfgfilenames.append(cfgfname)

        self.cfg = gc3libs.config.Configuration(
            cfgfilenames[0],
            self.cfg.cfgfiles[0],
            cfgfilenames[1], auto_enable_auth=True)

        self.core = gc3libs.core.Core(self.cfg)
        for resource in self.core.get_resources():
            for (k, v) in resource.items():
                if k not in ['prologue', 'epilogue',
                             'myapp_prologue', 'myapp_epilogue']:
                    continue
                assert os.path.isabs(v)
                assert os.path.isfile(v)
                assert (os.path.abspath(v) ==
                             v)


def test_invalid_resource_type():
    """Test parsing a configuration file with an unknown resource type."""
    tmpfile = _setup_config_file("""
[resource/test]
type = invalid
auth = none
transport = local
max_cores_per_job = 1
max_memory_per_core = 1
max_walltime = 8
max_cores = 2
architecture = x86_64
    """)
    try:
        cfg = gc3libs.config.Configuration(tmpfile)
        with pytest.raises(gc3libs.exceptions.ConfigurationError):
            # pylint: disable=unused-variable
            resources = cfg.make_resources(ignore_errors=False)
    finally:
        os.remove(tmpfile)


def test_removed_arc0_resource_type():
    """Test parsing a configuration file with the removed `arc0` resource type."""
    _test_removed_resource_type('arc0')

def test_removed_arc1_resource_type():
    """Test parsing a configuration file with the removed `arc1` resource type."""
    _test_removed_resource_type('arc1')

def _test_removed_resource_type(kind):
    name = ('test_' + kind)
    tmpfile = _setup_config_file("""
[resource/{name}]
type = {kind}
auth = none
transport = local
max_cores_per_job = 11
max_memory_per_core = 22 GB
max_walltime = 33 hours
max_cores = 44
architecture = x86_64
    """.format(name=name, kind=kind))
    try:
        cfg = gc3libs.config.Configuration(tmpfile)
        resources = cfg.make_resources(ignore_errors=False)
        assert name not in resources
    finally:
        os.remove(tmpfile)


def test_additional_backend():
    """Test instanciating a non-std backend."""
    tmpfile = _setup_config_file("""
[resource/test]
type = noop
auth = none
transport = local
max_cores_per_job = 1
max_memory_per_core = 1
max_walltime = 8
max_cores = 2
architecture = x86_64
    """)
    try:
        cfg = gc3libs.config.Configuration(tmpfile)
        cfg.TYPE_CONSTRUCTOR_MAP['noop'] = ('gc3libs.backends.noop', 'NoOpLrms')
        resources = cfg.make_resources(ignore_errors=False)
        # resources are enabled by default
        assert 'test' in resources
        from gc3libs.backends.noop import NoOpLrms
        assert isinstance(resources['test'], NoOpLrms)
        # test types
    finally:
        # since TYPE_CONSTRUCTOR_MAP is a class-level variable, we
        # need to clean up otherwise other tests will see the No-Op
        # backend
        del cfg.TYPE_CONSTRUCTOR_MAP['noop']
        os.remove(tmpfile)


def test_resource_definition_via_dict():
    """Test programmatic resource definition (as opposed to reading a config file)."""
    cfg = gc3libs.config.Configuration()
    # define resource
    name = 'test'
    cfg.resources[name].update(
        name=name,
        type='shellcmd',
        auth='none',
        transport='local',
        max_cores_per_job=1,
        max_memory_per_core=1*GB,
        max_walltime=8*hours,
        max_cores=2,
        architecture=Run.Arch.X86_64,
    )
    # make it
    resources = cfg.make_resources(ignore_errors=False)
    # check result
    resource = resources[name]
    assert resource.name == name
    assert isinstance(resource,
                       gc3libs.backends.shellcmd.ShellcmdLrms)
    assert isinstance(resource.transport,
                       gc3libs.backends.transport.LocalTransport)
    assert resource.max_cores_per_job == 1
    assert resource.max_memory_per_core == 1*GB
    assert resource.max_walltime == 8*hours
    assert resource.max_cores == 2
    assert resource.architecture == Run.Arch.X86_64


def test_multiple_instanciation(num_resources=3):
    """Check that multiple resources of the same type can be instanciated."""
    cfg = gc3libs.config.Configuration()
    for n in range(num_resources):
        name = ('test%d' % (n+1))
        cfg.resources[name].update(
            name=name,
            type='shellcmd',
            auth='none',
            transport='local',
            max_cores_per_job=1,
            max_memory_per_core=1*GB,
            max_walltime=8*hours,
            max_cores=2,
            architecture=Run.Arch.X86_64,
        )
    resources = cfg.make_resources(ignore_errors=False)
    for n in range(num_resources):
        name = ('test%d' % (n+1))
        resource = resources[name]
        assert resource.name == name
        assert isinstance(resource, gc3libs.backends.shellcmd.ShellcmdLrms)


def test_large_file_settings():
    """Check that 'large file' SSH settings are correctly parsed"""
    tmpfile = _setup_config_file("""
[auth/ssh]
type = ssh
username = gc3pie

[resource/test]
type = shellcmd
auth = ssh
transport = ssh
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8
max_cores = 2
architecture = x86_64
override = no
large_file_threshold = 1 GB
large_file_chunk_size = 200 MB
    """)
    try:
        cfg = gc3libs.config.Configuration(tmpfile)
        resources = cfg.make_resources(ignore_errors=False)
        # resources are enabled by default
        assert 'test' in resources
        backend = resources['test']
        assert isinstance(backend, ShellcmdLrms)
        assert isinstance(backend.transport, SshTransport)
        assert backend.transport.large_file_threshold == 10**9 # 1*GB in bytes
        assert backend.transport.large_file_chunk_size == 200*10**6 # 200*MB in bytes
    finally:
        os.remove(tmpfile)


def test_override_config_file_location():
    """Check that config file location is overriden by env var GC3PIE_CONF"""
    tmpfile = _setup_config_file("""
[resource/test]
type = shellcmd
auth = none
transport = local
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8h
max_cores = 2
architecture = x86_64
override = False
    """)
    gc3pie_conf_orig = os.environ.get('GC3PIE_CONF')
    try:
        os.environ['GC3PIE_CONF'] = tmpfile
        from gc3libs import defaults
        try:
            reload(defaults)  # Py2
        except NameError:
            import importlib; importlib.reload(defaults)  # Py3
        with mock.patch('gc3libs.config.open') as mo:
            mo.return_value = open(tmpfile, 'r')
            gc3libs.create_core()
            mo.assert_called_with(tmpfile, 'r')
    finally:
        os.remove(tmpfile)
        if gc3pie_conf_orig is None:
            del os.environ['GC3PIE_CONF']
        else:
            os.environ['GC3PIE_CONF'] = gc3pie_conf_orig


def test_override_logging_config_location():
    """Check that logging config file location is overriden by env var GC3PIE_CONF"""
    gc3pie_conf_orig = os.environ.get('GC3PIE_CONF')
    with temporary_directory() as tmpdir:
        # create config file
        cfg_loc = os.path.join(tmpdir, 'gc3pie.conf')
        with open(cfg_loc, 'w') as cfg_file:
            cfg_file.write('# actually ignored')
        # create logging config file
        logcfg_loc = os.path.join(tmpdir, 'gc3pie.log.conf')
        with open(logcfg_loc, 'w') as logcfg_file:
            # logging.fileConfig() errors out if not given a valid
            # file with all the sections and values...
            logcfg_file.write('''
[formatters]
keys=none

[formatter_none]

[handlers]
keys=none

[handler_none]
class=NullHandler
args=()

[loggers]
keys=root

[logger_root]
level=NOTSET
handlers=none
''')
        # point to alternate config file
        os.environ['GC3PIE_CONF'] = cfg_loc
        # do the actual check
        #
        # `logging.fileConfig()` module does not perform an `open()`
        # call itself -- rather it uses a `ConfigParser` object to
        # read the file; this means we have to patch `ConfigParser` to
        # intercept its calls to `open()`, which in turn means we need
        # to discriminate between Py2 and Py3 because that module was
        # renamed in the transition...
        try:
            import ConfigParser  # Py2
            open_fn = 'ConfigParser.open'
            extra_args = {}
        except ImportError:
            import configparser  # Py3
            open_fn = 'configparser.open'
            extra_args = {'encoding': None}
        with mock.patch(open_fn) as mo:
            mo.return_value = open(logcfg_loc)
            gc3libs.configure_logger()
            mo.assert_called_with(logcfg_loc, **extra_args)
    if gc3pie_conf_orig is None:
        del os.environ['GC3PIE_CONF']
    else:
        os.environ['GC3PIE_CONF'] = gc3pie_conf_orig


if __name__ == "__main__":
    import pytest
    pytest.main(["-v", __file__])



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2011-2014, 2018  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
__docformat__ = 'reStructuredText'

# stdlib imports
import os
import sys
from weakref import WeakValueDictionary

# GC3Pie imports
import gc3libs
import gc3libs.exceptions
from gc3libs.utils import same_docstring_as
from gc3libs.url import Url

from gc3libs.persistence.idfactory import IdFactory
from gc3libs.persistence.serialization import (DEFAULT_PROTOCOL, make_pickler,
                                               make_unpickler)
from gc3libs.persistence.store import Store


# persist objects in a filesystem directory

class FilesystemStore(Store):

    """
    Save and load objects in a given directory.  Uses Python's
    standard `pickle` module to serialize objects onto files.

    All objects are saved as files in the given directory (default:
    `gc3libs.defaults.JOBS_DIR`).  The file name is the object ID.

    If an object contains references to other `Persistable` objects,
    these are saved in the file they would have been saved if the
    `save` method was called on them in the first place, and only an
    'external reference' is saved in the pickled container. This
    ensures that: (1) only one copy of a shared object is ever saved,
    and (2) any shared reference to `Persistable` objects is correctly
    restored when restoring the container.

    The default `idfactory` assigns object IDs by appending a
    sequential number to the class name; see class `Id` for
    details.

    The `protocol` argument specifies the serialization protocol to use,
    if different from `gc3libs.persistence.serialization.DEFAULT_PROTOCOL`.

    Any extra keyword arguments are ignored for compatibility with
    `SqlStore`.
    """

    def __init__(self,
                 directory=gc3libs.defaults.JOBS_DIR,
                 idfactory=IdFactory(),
                 protocol=DEFAULT_PROTOCOL,
                 **extra_args):
        if isinstance(directory, Url):
            super(FilesystemStore, self).__init__(directory)
            directory = directory.path
        else:
            super(FilesystemStore, self).__init__(
                Url(scheme='file', path=os.path.abspath(directory)))
        self._directory = directory

        self.idfactory = idfactory
        self._loaded = WeakValueDictionary()
        self._protocol = protocol

    @same_docstring_as(Store.invalidate_cache)
    def invalidate_cache(self):
        self._loaded.clear()

    @same_docstring_as(Store.list)
    def list(self):
        if not os.path.exists(self._directory):
            return []
        return [id_ for id_ in os.listdir(self._directory)
                if not id_.endswith('.OLD')]

    def _load_from_file(self, path):
        """Auxiliary method for `load`."""
        # gc3libs.log.debug("Loading object from file '%s' ...", path)
        with open(path, 'rb') as src:
            unpickler = make_unpickler(self, src)
            obj = unpickler.load()
            return obj

    @same_docstring_as(Store.load)
    def load(self, id_):
        # return cached copy, if any
        try:
            return self._loaded[str(id_)]
        except KeyError:
            pass

        # no cached copy, load from disk
        filename = os.path.join(self._directory, id_)

        sources = [filename, filename + '.OLD']
        for source in sources:
            if not os.path.exists(source):
                gc3libs.log.debug(
                    "Cannot load object %s from '%s':"
                    " file does not exist", id_, source)
                continue
            try:
                obj = self._load_from_file(source)
                break  # exit `for source in sources` loop ...
            except Exception as ex:
                gc3libs.log.warning(
                    "Failed loading file '%s': %s: %s",
                    filename, ex.__class__.__name__, ex,
                    exc_info=True)
        else:
            # complain loudly
            raise gc3libs.exceptions.LoadError(
                "Failed loading object %s from file(s) %r."
                " (Earlier log lines may provide more details.)"
                % (id_, sources))

        # minimal sanity check
        if not hasattr(obj, 'persistent_id'):
            raise gc3libs.exceptions.LoadError(
                "Invalid format in file '%s':"
                " missing 'persistent_id' attribute"
                % (filename,))
        if str(obj.persistent_id) != str(id_):
            raise gc3libs.exceptions.LoadError(
                "Retrieved persistent ID '%s' %s"
                " does not match given ID '%s' %s"
                % (obj.persistent_id, type(obj.persistent_id),
                   id_, type(id_)))

        # maybe update object after GC3Pie update?
        super(FilesystemStore, self)._update_to_latest_schema()

        # update cache
        assert str(id_) not in self._loaded
        self._loaded[str(id_)] = obj

        return obj

    @same_docstring_as(Store.remove)
    def remove(self, id_):
        filename = os.path.join(self._directory, id_)
        os.remove(filename)
        try:
            del self._loaded[str(id_)]
        except KeyError:
            pass

    @same_docstring_as(Store.replace)
    def replace(self, id_, obj):
        self._save_or_replace(id_, obj)

    @same_docstring_as(Store.save)
    def save(self, obj):
        if not hasattr(obj, 'persistent_id'):
            obj.persistent_id = self.idfactory.new(obj)
        self._save_or_replace(obj.persistent_id, obj)
        return obj.persistent_id

    def _save_or_replace(self, id_, obj):
        """
        Save `obj` into file identified by `id_`; if no such
        destination file exists, create it.  Ensure that the
        destination file is kept intact in case dumping `obj` fails.
        """
        filename = os.path.join(self._directory, id_)
        # gc3libs.log.debug("Storing job '%s' into file '%s'", obj, filename)

        if not os.path.exists(self._directory):
            try:
                os.makedirs(self._directory)
            except Exception as ex:
                # raise same exception but add context message
                gc3libs.log.error("Could not create jobs directory '%s': %s"
                                  % (self._directory, str(ex)))
                raise

        backup = None
        if os.path.exists(filename):
            backup = filename + '.OLD'
            os.rename(filename, backup)

        with open(filename, 'w+b') as tgt:
            try:
                pickler = make_pickler(self, tgt, obj)
                pickler.dump(obj)
            except Exception as err:
                gc3libs.log.error(
                    "Error saving task '%s' to file '%s': %s: %s",
                    obj, filename, err.__class__.__name__, err)
                # move backup file back in place
                if backup is not None:
                    try:
                        os.rename(backup, filename)
                    except:
                        pass  # ignore errors
                raise
            if hasattr(obj, 'changed'):
                obj.changed = False
            # remove backup file, if exists
            try:
                os.remove(backup)
            except:
                pass  # ignore errors
            # update cache
            if id_ in self._loaded:
                old = self._loaded[str(id_)]
                if old is not obj:
                    self._loaded[str(id_)] = obj


def make_filesystemstore(url, *args, **extra_args):
    """
    Return a `FilesystemStore`:class: instance, given a 'file:///' URL
    and optional initialization arguments.

    This function is a bridge between the generic factory functions
    provided by `gc3libs.persistence.make_store`:func: and
    `gc3libs.persistence.register`:func: and the class constructor
    `FilesystemStore`:class.

    Examples::

      >>> fs1 = make_filesystemstore(Url('file:///tmp'))
      >>> fs1.__class__.__name__
      'FilesystemStore'
    """
    assert isinstance(url, Url)
    return FilesystemStore(url.path, *args, **extra_args)


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="filesystem",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2011, 2012, 2019,  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
__docformat__ = 'reStructuredText'


import pytest

from gc3libs import Application
import gc3libs.exceptions


def test_invalid_invocation():
    with pytest.raises(TypeError):
        Application()

app_mandatory_arguments = (
    'arguments',
    'inputs',
    'outputs',
    'output_dir',
)

@pytest.mark.parametrize("mandatory", app_mandatory_arguments)
def test_mandatory_arguments(mandatory):
    # check for all mandatory arguments
    args = {
        'arguments': ['/bin/true'],
        'inputs': [],
        'outputs': [],
        'output_dir': '/tmp',
    }

    # test *valid* invocation
    Application(**args)

    del args[mandatory]

    # now check that invocation is *invalid*
    with pytest.raises(TypeError):
        Application(**args)


app_wrong_arguments = (
    # 'inputs' : ['duplicated', 'duplicated'],
    # duplicated inputs doesnt raise an exception but just a warning
    ('outputs', ['/should/not/be/absolute']),
    # 'outputs' : ['duplicated', 'duplicated'],
    # duplicated outputs doesnt raise an exception but just a warning
    ('requested_architecture', 'FooBar'),
    ('requested_cores', 'one'),
)

@pytest.mark.parametrize("wrongarg", app_wrong_arguments)
def test_wrong_type_arguments(wrongarg):
    # Things that will raise errors:
    # * unicode arguments
    # * unicode files in inputs or outputs
    # * remote paths (outputs) must not be absolute
    #
    # What happens when you request non-integer cores/memory/walltime?
    # what happens when you request non-existent architecture?

    args = {
        'arguments': ['/bin/true'],
        'inputs': [],
        'outputs': [],
        'output_dir': '/tmp',
        'requested_cores': 1,
    }

    key, value = wrongarg

    args[key] = value
    with pytest.raises((gc3libs.exceptions.InvalidArgument, ValueError)):
        Application(**args)


def test_valid_invocation():
    ma = {'arguments': ['/bin/true'],
          'inputs': ['/tmp/a', 'b'],
          'outputs': ['o1', 'o2'],
          'output_dir': '/tmp',
          }
    Application(**ma)


def test_io_spec_to_dict_mixed():
    # pylint: disable=import-error,protected-access,redefined-outer-name
    from gc3libs.utils import text_str
    d = Application._io_spec_to_dict(
          gc3libs.url.UrlKeyDict, {
              u'/tmp/\u0246': u'\u0246',
              b'/tmp/b/': b'b'},
        True)
    for k, v in d.items():
        # Python 3+
        assert isinstance(v, text_str)
        assert isinstance(k.path, text_str)


# main: run tests

if "__main__" == __name__:
    pytest.main(["-v", __file__])



#! /usr/bin/env python
#
"""
Driver script for running the `forwardPremium` application on SMSCG.
"""
# Copyright (C) 2011  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__author__ = 'Benjamin Jonen <benjamin.jonen@bf.uzh.ch'
# summary of user-visible changes
__changelog__ = """

"""
__docformat__ = 'reStructuredText'

from __future__ import absolute_import, print_function
import os, sys
import numpy as np

# import personal libraries
path2Pymods = os.path.join(os.path.dirname(__file__), '../')
if not sys.path.count(path2Pymods):
    sys.path.append(path2Pymods)
path2Src = os.path.join(os.path.dirname(__file__), '../src')
if not sys.path.count(path2Src):
    sys.path.append(path2Src)

# Remove all files in curPath if -N option specified. 
if __name__ == '__main__':    
    import sys
    if '-N' in sys.argv:
        import os, shutil
        path2Pymods = os.path.join(os.path.dirname(__file__), '../')
        if not sys.path.count(path2Pymods):
            sys.path.append(path2Pymods)
        curPath = os.getcwd()
        filesAndFolder = os.listdir(curPath)
        if 'taskColAppPot.csv' in filesAndFolder: # if another paraSearch was run in here before, clean up. 
            if 'para.loop' in os.listdir(os.getcwd()):
                shutil.copyfile(os.path.join(curPath, 'para.loop'), os.path.join('/tmp', 'para.loop'))
                shutil.rmtree(curPath)
                shutil.copyfile(os.path.join('/tmp', 'para.loop'), os.path.join(curPath, 'para.loop'))
                os.remove(os.path.join('/tmp', 'para.loop'))
            else:
                shutil.rmtree(curPath)

if __name__ == "__main__":
    import taskColAppPot


# superclasses
from idRisk import idRiskApplication, idRiskApppotApplication
from paraLoop import paraLoop


# gc3 library imports
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, existing_file
from gc3libs.workflow import SequentialTaskCollection, ParallelTaskCollection
import gc3libs.utils

import gc3libs.application.apppot

import gc3libs.debug

# logger
from pymods.support.support import fillInputDir
from pymods.support.support import wrapLogger
import pymods.support.support as support
from pymods.classes.tableDict import tableDict

# Set up logger
logger = wrapLogger(loggerName = 'idRiskParaSearchLogger', streamVerb = 'DEBUG', logFile = os.path.join(os.getcwd(), 'idRiskParaSearch.log'))

# call -x /home/benjamin/workspace/idrisk/model/bin/idRiskOut -b /home/benjamin/workspace/idrisk/model/base para.loop -C 10 -N -A '/home/benjamin/apppot0+ben.diskUpd.img'

class solveParaCombination(SequentialTaskCollection):

    def __init__(self, substs, solverParas, **sessionParas):

        logger.debug('entering solveParaCombination.__init__ for job %s' % sessionParas['jobname'])
        self.iter    = 0

        self.jobname = 'solverParacombination' + sessionParas['jobname']
        self.substs = substs

        self.sessionParas     = sessionParas
        self.pathToExecutable = sessionParas['pathToExecutable']
        self.architecture     = sessionParas['architecture']
        self.localBaseDir     = sessionParas['localBaseDir']
        
        self.paraFolder = os.path.join(os.getcwd(), sessionParas['jobname'])
        
        self.wBarLower_task = idRiskParaSearchDriver(self.paraFolder, self.substs, solverParas, **sessionParas)

        SequentialTaskCollection.__init__(self, self.jobname, [self.wBarLower_task])

        logger.debug('done gParaSearchDriver.__init__ for job %s' % sessionParas['jobname'])

    def __str__(self):
        return self.jobname

    def next(self, *args): 
        self.iter += 1
        if self.wBarLower_task.execution.returncode == 13:
            logger.critical('wBarLower failed. terminating para combo')
            self.execution.returncode = 13
            return Run.State.TERMINATED
        elif self.wBarLower_task.execution.returncode == 0:
            #wBarTable = tableDict.fromTextFile(os.path.join(self.paraFolder, 'optimwBarLower', 'overviewSimu'), width = 20, prec = 10)
            #optimalRunTable = wBarTable.getSubset( np.abs(wBarTable['wBarLower'] - self.wBarLower_task.costlyOptimizer.best_x) < 1.e-7 )
            #optimalRunFile = open(os.path.join(self.paraFolder, 'optimalRun'), 'w')
            #print >> optimalRunFile, optimalRunTable
            self.execution.returncode = 0
            return Run.State.TERMINATED
        else:
            print 'unknown return code'
            os._exit

class idRiskParaSearchDriver(SequentialTaskCollection):

    def __init__(self, paraFolder, substs, solverParas, **sessionParas):

        logger.debug('entering solveParaCombination.__init__ for job %s' % sessionParas['jobname'])
        self.iter    = 0

        self.jobname = 'idRiskParaSearchDriver' + sessionParas['jobname']
        self.substs = substs

        self.sessionParas     = sessionParas
        self.pathToExecutable = sessionParas['pathToExecutable']
        self.architecture     = sessionParas['architecture']
        self.localBaseDir     = sessionParas['localBaseDir']
        
        self.paraFolder = os.path.join(os.getcwd(), sessionParas['jobname'])
        # setup AppPot parameters
        use_apppot = False
        apppot_img = None  
        apppot_changes = None
        apppot_file = sessionParas['AppPotFile']
        if apppot_file:
            use_apppot = True
            if apppot_file.endswith('.changes.tar.gz'):
                apppot_changes = apppot_file
            else:
                apppot_img = apppot_file
        pathToExecutable = sessionParas['pathToExecutable']
        localBaseDir     = sessionParas['localBaseDir']
        architecture     = sessionParas['architecture']
        jobname          = sessionParas['jobname']
        executable = os.path.basename(self.pathToExecutable)
        inputs = { self.pathToExecutable:executable }
        # make a "stage" directory where input files are collected
        path_to_stage_dir = os.path.join(self.paraFolder, jobname)
        gc3libs.utils.mkdir(path_to_stage_dir)
        prefix_len = len(path_to_stage_dir) + 1
        # 2. apply substitutions to parameter files
        for (path, changes) in substs.iteritems():
            for (var, val, index, regex) in changes:
                support.update_parameter_in_file(os.path.join(localBaseDir, path),
                                         var, index, val, regex)
        support.fillInputDir(localBaseDir, path_to_stage_dir)
        # 3. build input file list
        for dirpath,dirnames,filenames in os.walk(path_to_stage_dir):
            for filename in filenames:
                # cut the leading part, which is == to path_to_stage_dir
                relpath = dirpath[prefix_len:]
                # ignore output directory contents in resubmission
                if relpath. startswith('output'):
                    continue
                remote_path = os.path.join(relpath, filename)
                inputs[os.path.join(dirpath, filename)] = remote_path
        # all contents of the `output` directory are to be fetched
        outputs = { 'output/':'' }
        kwargs = {}
        kwargs['stdout'] = 'idRisk.log'
        kwargs['join'] = True
        kwargs['output_dir'] = os.path.join(path_to_stage_dir, 'output')
        kwargs['requested_architecture'] = self.architecture
        
        # adaptions for uml
        if use_apppot:
            if apppot_img is not None:
                kwargs['apppot_img'] = apppot_img
            if apppot_changes is not None:
                kwargs['apppot_changes'] = apppot_changes
            cls = idRiskApppotApplication
        else:
            cls = idRiskApplication 
        kwargs.setdefault('tags', [ ])

        # hand over job to create
#        self.curApplication = cls('/home/user/job/' + executable, [], inputs, outputs, **kwargs)
        self.curApplication = cls('./' + executable, [], inputs, outputs, **kwargs)

        SequentialTaskCollection.__init__(self, self.jobname, [ self.curApplication ])

        logger.debug('done gParaSearchDriver.__init__ for job %s' % sessionParas['jobname'])

    def __str__(self):
        return self.jobname

    def next(self, *args): 
        self.iter += 1
        if self.curApplication.execution.returncode == 13:
            logger.critical('wBarLower failed. terminating para combo')
            self.execution.returncode = 13
            return Run.State.TERMINATED
        elif self.curApplication.execution.returncode == 0:            
            self.execution.returncode = 0
            return Run.State.TERMINATED
        else:
            print 'unknown return code'
            os._exit


class taskColAppPotScript(SessionBasedScript, paraLoop):
    """
      Read `.loop` files and execute the `forwardPremium` program accordingly.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = '0.2',
            # only '.loop' files are considered as valid input
            input_filename_pattern = '*.loop',
        )
        paraLoop.__init__(self, 'INFO')        

    def setup_options(self):
        self.add_param("-b", "--initial", metavar="DIR",
                       dest="initial",
                       help="Include directory contents in any job's input."
                       " Use this to specify the initial guess files.")
        self.add_param("-n", "--dry-run",
                       dest = 'dryrun', action="store_true", default = False,
                       help="Take the loop for a test run")
        self.add_param("-x", "--executable", metavar="PATH",
                       dest="executable", default=os.path.join(
                           os.getcwd(), "forwardPremiumOut"),
                       help="Path to the `forwardPremium` executable binary"
                       "(Default: %(default)s)")
        self.add_param("-X", "--architecture", metavar="ARCH",
                       dest="architecture", default=Run.Arch.X86_64,
                       help="Processor architecture required by the executable"
                       " (one of: 'i686' or 'x86_64', without quotes)")
        self.add_param("-A", "--apppot", metavar="PATH",
                       dest="apppot",
                       type=existing_file, default=None,
                       help="Use an AppPot image to run idRisk."
                       " PATH can point either to a complete AppPot system image"
                       " file, or to a `.changes` file generated with the"
                       " `apppot-snap` utility.")
        self.add_param("-mP", "--makePlots", metavar="ARCH", type = bool, 
                       dest="makePlots", default = True,
                       help="Generate population plots each iteration.  ")
        self.add_param("-xVars", "--xVars", metavar="ARCH",
                       dest="xVars", default = 'EA',
                       help="x variables over which to optimize")
        self.add_param("-xVarsDom", "--xVarsDom", metavar="ARCH",
                       dest="xVarsDom", default = '0.5 0.9',
                       help="Domain to sample x values from. Space separated list. ")
        self.add_param("-targetVars", "--target_fx", metavar="ARCH",
                       dest="targetVars", default = '0.1',
                       help="Domain to sample x values from. Space separated list. ")
        self.add_param("-target_fx", "--target_fx", metavar="ARCH",
                       dest="target_fx", default = '0.1',
                       help="Domain to sample x values from. Space separated list. ")
        self.add_param("-sv", "--solverVerb", metavar="ARCH",
                       dest="solverVerb", default = 'DEBUG',
                       help="Separate verbosity level for the global optimizer ")
        self.add_param("-yC", "--yConvCrit", metavar="ARCH", type = float, 
                       dest="convCrit", default = '1.e-2',
                       help="Convergence criteria for y variables. ")

    def parse_args(self):
        """
        Check validity and consistency of command-line options.
        """
        if not os.path.exists(self.params.executable):
            raise gc3libs.exceptions.InvalidUsage(
                "Path '%s' to the 'forwardPremium' executable does not exist;"
                " use the '-x' option to specify a valid one."
                % self.params.executable)
        if os.path.isdir(self.params.executable):
            self.params.executable = os.path.join(self.params.executable,
                                                  'forwardPremium')
        gc3libs.utils.check_file_access(self.params.executable, os.R_OK|os.X_OK,
                                gc3libs.exceptions.InvalidUsage)



    def new_tasks(self, extra):     

        paraLoopFile = self._search_for_input_files(self.params.args).pop()  # search_... returns set.

        # Copy base dir
        localBaseDir = os.path.join(os.getcwd(), 'localBaseDir')
        shutil.copytree(self.params.initial, localBaseDir)

        for jobname, substs in self.process_para_file(paraLoopFile):
            # yield job
            sessionParas = {}
            sessionParas['AppPotFile'] = self.params.apppot
            sessionParas['pathToExecutable'] = self.params.executable
            sessionParas['architecture'] = self.params.architecture
            sessionParas['localBaseDir'] = localBaseDir
            sessionParas['jobname'] = jobname
            # Compute domain
            xVarsDom = self.params.xVarsDom.split()
            lowerBds = np.array([xVarsDom[i] for i in range(len(xVarsDom)) if i % 2 == 0], dtype = 'float64')
            upperBds = np.array([xVarsDom[i] for i in range(len(xVarsDom)) if i % 2 == 1], dtype = 'float64')
            domain = zip(lowerBds, upperBds)
            solverParas = {}
            solverParas['xVars'] = self.params.xVars.split()
            solverParas['xInitialParaCombo'] = np.array([lowerBds, upperBds])
            solverParas['targetVars'] = self.params.xVars.split()
            solverParas['target_fx'] = map(float, self.params.target_fx.split())
            solverParas['plotting'] = False
            solverParas['convCrit'] = self.params.convCrit
            yield (jobname, solveParaCombination, [ substs, solverParas ], sessionParas)
            
            
if __name__ == '__main__':
    logger.info('Starting: \n%s' % ' '.join(sys.argv))
    taskColAppPotScript().run()
    logger.info('main done')
logger.info('done')


#! /usr/bin/env python

"""
Support for communication between parts of code through "events".

This file collects definitions of event classes used across the
library code. The actual subscription and notification mechanisms
come from the implementation of the `Observable/Observer pattern`__
provided by Python's library `generic`__.

.. __: https://en.wikipedia.org/wiki/Observer_pattern
.. __: http://generic.readthedocs.io/en/latest/event_system.html#event-system
"""

# Copyright (C) 2018  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
__docformat__ = 'reStructuredText'


# do not make symbols imported from `blinker` public: use of `blinker`
# here is an implementation detail
from blinker import signal as _signal


TaskStateChange = _signal('task_state_change')

TermStatusChange = _signal('task_termstatus_change')



#! /usr/bin/env python

"""
"""

# Copyright (C) 2012-2014  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
from builtins import object
__docformat__ = 'reStructuredText'

import os

import gc3libs
from gc3libs.authentication import Auth
import gc3libs.exceptions


class OpenStackAuth(object):

    def __init__(self, **auth):
        try:
            # test validity
            assert auth['type'] == 'openstack', \
                "Configuration error. Unknown type; %s. " \
                "Valid type: openstack" \
                % auth.type
        except AssertionError as x:
            raise gc3libs.exceptions.ConfigurationError(
                'Erroneous configuration parameter: %s' % str(x))

        try:
            for (key, var) in (
                    # list of mandatory env vars taken from:
                    # http://docs.openstack.org/user-guide/common/cli_set_environment_variables_using_openstack_rc.html
                    ('os_auth_url',     'OS_AUTH_URL'),
                    ('os_password',     'OS_PASSWORD'),
                    ('os_project_name', 'OS_TENANT_NAME'),
                    ('os_username',     'OS_USERNAME'),
            ):
                if key not in auth:
                    auth[key] = os.getenv(var)
                    assert auth[key], (
                        "Missing mandatory configuration parameter for {name} auth:"
                        " either define the `{key}` configuration key,"
                        " or set the `{var}` environmental variable."
                        .format(name=("'" + auth['name'] + "'"), key=key, var=var,))
                    # Strip quotes from os_* in case someone put
                    # it in the configuration file
                    auth[key] = auth[key].strip('"').strip("'")
        except AssertionError as x:
            raise gc3libs.exceptions.ConfigurationError(str(x))

        self.__dict__.update(auth)

    def check(self):
        return True

    def enable(self):
        return True


Auth.register('openstack', OpenStackAuth)

# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="openstack",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   gbugs.py -- Front-end script for running MCMC with opneBUGS
#   function throguh R over a large table set.
#
#   Copyright (C) 2011, 2012  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `R` jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gbugs.py --help`` for program usage
instructions.

Input parameters consists of:
:param str table input file: Path to an .txt file containing input data in table form

"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2015-04-10:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gbugs
    gbugs.GBugsScript().run()

import os
import sys
import time
import tempfile

import shutil
# import csv

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask, StagedTaskCollection, ParallelTaskCollection

DEFAULT_CORES = 1
DEFAULT_MEMORY = Memory(1500,MB)
DEFAULT_WALLTIME = Duration(300,hours)

def generate_chunked_files_and_list(file_to_chunk, chunk_size=1000, tmp_folder=os.getcwd()):
        """
        Takes a file_name as input and a defined chunk size
        ( uses 1000 as default )
        returns a list of filenames 1 for each chunk created and a corresponding
        index reference
        e.g. ('/tmp/chunk2800.csv,2800) for the chunk segment that goes
        from 2800 to (2800 + chunk_size)
        """
        index = 0
        chunk = []
        failure = False
        chunk_files_dir = os.path.join(tmp_folder,"tmp")

        # creating 'chunk_files_dir'
        if not(os.path.isdir(chunk_files_dir)):
            try:
                os.mkdir(chunk_files_dir)
            except OSError, osx:
                gc3libs.log.error("Failed while creating tmp folder %s. " % chunk_files_dir +
                                  "Error %s." % str(osx) +
                                  "Using default '/tmp'")
                chunk_files_dir = "/tmp"

        try:
            fd = open(file_to_chunk,'rb')

            fout = None

            for (i, line) in enumerate(fd):
                if i % chunk_size == 0:
                    if fout:
                        fout.close()
                    (handle, tmp_filename) = tempfile.mkstemp(dir=chunk_files_dir,
                                                                    prefix=
                                                                   'gbugs-',
                                                                   suffix=
                                                                   "%d.txt" % i)
                    fout = open(tmp_filename,'w')
                    chunk.append((fout.name,i))
                fout.write(line)
            fout.close()
        except OSError, osx:
            gc3libs.log.critical("Failed while creating chunk files." +
                                 "Error %s", (str(osx)))
            failure = True
        finally:
            if failure:
                # remove all tmp file created
                gc3libs.log.info("Could not generate full chunk list. "
                                 "Removing all existing tmp files... ")
                for (cfile, index) in chunk:
                    try:
                        os.remove(cfile)
                    except OSError, osx:
                        gc3libs.log.error("Failed while removing " +
                                          "tmp file %s. " +
                                          "Message %s" % osx.message)

        return chunk


## custom application class
class GBugsApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'gbugs'

    def __init__(self, input_table_filename, **extra_args):

        # setup output references
        # self.result_dir = result_dir
        self.output_dir = extra_args['output_dir']
        self.output_filename = 'result-%s.Rdata' % extra_args['index_chunk']
        self.output_file = os.path.join(self.output_dir,self.output_filename)
        outputs = [("./output.Rdata",self.output_filename)]
        # setup input references
        inputs = dict()
        inputs[input_table_filename] = "./input.txt"
        arguments = "Rscript --vanilla "

        # check the optional inputs

        if extra_args.has_key('driver_script'):
            inputs[extra_args['driver_script']] = "./run.R"

        arguments +=  "run.R "
        arguments += "input.txt output.Rdata"

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gbugs.log',
            join=True,
            **extra_args)

    def terminated(self):
        """
        Check whether output file has been properly created
        """
        # XXX: TBD, work on a more precise checking of the output file
        gc3libs.log.info("Application terminated with exit code %s" % self.execution.exitcode)
        if (not os.path.isfile(self.output_file)):
            gc3libs.log.error("Failed while checking outputfile %s." % self.output_file)
            self.execution.returncode = (0, 99)

class OpenBUGSCollection(StagedTaskCollection):
    def __init__(self, input_table_file, chunk_size, driver_script, stats_script, **extra_args):
        self.name = os.path.basename(input_table_file)
        self.input_table_file = input_table_file
        self.chunk_size = chunk_size
        self.driver_script = driver_script
        self.stats_script = stats_script
        self.output_dir = extra_args['output_dir']
        self.extra = extra_args
        StagedTaskCollection.__init__(self)

    def stage0(self):
        """
        Chunk input table and run chunks in parallel
        """
        tasks = []
        for (input_file, index_chunk) in generate_chunked_files_and_list(self.input_table_file,
                                                                              self.chunk_size):
            jobname = "gbugs-%s" % (str(index_chunk))
            extra_args = self.extra.copy()
            extra_args['index_chunk'] = str(index_chunk)
            extra_args['jobname'] = jobname

            # extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)

            if self.driver_script:
                extra_args['driver_script'] = self.driver_script

            gc3libs.log.debug("Creating Task for index : %d - %d" %
                           (index_chunk, (index_chunk + self.chunk_size)))

            tasks.append(GBugsApplication(
                    input_file,
                    **extra_args))
        return ParallelTaskCollection(tasks)

class GBugsScript(SessionBasedScript):
    """
    Splits input .csv file into smaller chunks, each of them of size
    'self.params.chunk_size'.
    Then it submits one execution for each of the created chunked files.

    The ``gbugs`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gbugs`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gbugs``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = GBugsApplication,
            stats_only_for = GBugsApplication,
            )

    def setup_options(self):
        self.add_param("-k", "--chunk", metavar="[NUM]",
                       dest="chunk_size", default="1000",
                       help="How to split the edges input data set.")

        self.add_param("-M", "--master", metavar="[PATH]",
                       dest="driver_script", default=None,
                       help="Location of master driver R script.")

        self.add_param("-S", "--statistic", metavar="[PATH]",
                       dest="stats_script", default=None,
                       help="Location of statistic R script.")
    def setup_args(self):

        self.add_param('input_table', type=str,
                       help="Input input table full path name.")

    def parse_args(self):
        """
        Check presence of input folder (should contains R scripts).
        path to command_file should also be valid.
        """

        if not os.path.isfile(self.params.input_table):
            raise gc3libs.exceptions.InvalidUsage(
                "Invalid path to input table: '%s'. File not found"
                % self.params.input_table)

        self.table_filename = os.path.basename(self.params.input_table)

        # Verify that 'self.params.chunk_size' is int
        int(self.params.chunk_size)
        self.params.chunk_size = int(self.params.chunk_size)

    def new_tasks(self, extra):
        """
        Chunk initial input file
        For each chunked fule, generate a new GbugsTask
        """
        extra_args = extra.copy()
        extra_args['output_dir'] = self.params.output
        return [OpenBUGSCollection(self.params.input_table,
                                   self.params.chunk_size,
                                   self.params.driver_script,
                                   self.params.stats_script,
                                   **extra_args)]



#! /usr/bin/env python
#
#   gcelljunction.py -- GC3Pie front-end for running the
#   "tricellular_junction" code by T. Aegerter
#
#   Copyright (C) 2014, 2019  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for running multiple `tricellular_junction` instances.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gcelljunction --help`` for program usage instructions.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
* 2014-08-15: Add option to restart jobs from saved data.
* 2014-07-17: Snapshot output of RUNNING jobs at every cycle.
* 2014-03-10: Report on task progress in the "info" line.
* 2014-03-03: Initial release, forked off the ``gmhc_coev`` sources.
"""
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == '__main__':
    import gcelljunction
    gcelljunction.GCellJunctionScript().run()


# std module imports
import csv
import glob
import os
import re
import sys
import time
from pkg_resources import Requirement, resource_filename

# gc3 library imports
import gc3libs
import gc3libs.defaults
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript
from collections import defaultdict
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.utils import string_to_boolean
from gc3libs.workflow import RetryableTask


## custom application class

class GCellJunctionApplication(Application):
    """
    Custom class to wrap the execution of the ``tricellular_junction``
    program by T. Aegerter.
    """

    application_name = 'tricellular_junction'

    def __init__(self, sim_no, executable=None, restart=None, **extra_args):
        self.sim_no = sim_no
        wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                       "gc3libs/etc/gcelljunction_wrapper.sh")
        inputs = { wrapper_sh:os.path.basename(wrapper_sh) }
        extra_args.setdefault('requested_cores',        1)
        extra_args.setdefault('requested_memory',       3*GB)
        extra_args.setdefault('requested_architecture', Run.Arch.X86_64)
        extra_args.setdefault('requested_walltime',     60*Duration.days)
        # command-line parameters to pass to the tricellular_junction_* program
        self.sim_no = sim_no
        program_opts = [ sim_no ]
        if executable is not None:
            # use the specified executable
            exename = os.path.basename(executable)
            executable_name = './' + exename
            inputs[executable] = exename
            exe_opts = ['-x', executable_name]
        else:
            # assume one is installed in the VM
            executable_name = 'tricellular_junctions'
            exe_opts = [ ]
        self.restart = restart
        if self.restart:
            assert len(self.restart) == 3
            program_opts += [ 1, self.restart[0] ]
            inputs[self.restart[1]] = 'restart_data.mat'
            inputs[self.restart[2]] = 'restart_data4.mat'
        else:
            program_opts += [ 0, 0 ]
        Application.__init__(
            self,
            arguments=['./' + os.path.basename(wrapper_sh)] + exe_opts + [ '--' ] + program_opts,
            inputs = inputs,
            outputs = gc3libs.ANY_OUTPUT,
            stdout = 'tricellular_junctions.log',
            join=True,
            **extra_args)

    def terminated(self):
        restart = self.find_restart_data(self.output_dir)
        if restart:
            self.restart = restart

    @staticmethod
    def find_restart_data(output_dir):
        data_dir = os.path.join(output_dir, 'data')
        if os.path.isdir(data_dir):
            # find the latest output file
            entries = [ name
                        for name in os.listdir(data_dir)
                        if name.startswith('junction_') and name.endswith('.mat') ]
            latest = None
            latest_idx = 0
            for name in entries:
                idx = int(name[len('junction_'):-len('.mat')])
                if (latest is None) or (idx > latest_idx):
                    latest = name
                    latest_idx = idx
            assert latest is not None
            restart_data_file = os.path.join(data_dir, latest)
            restart_data_seqno = latest_idx
        else:
            restart_data_file = None
            restart_data_seqno = None

        data4_dir = os.path.join(output_dir, 'data4')
        if os.path.isdir(data4_dir):
            restart_data4_file = os.path.join(data4_dir, 'junctions.mat')
            if not os.path.exists(restart_data4_file):
                restart_data4_file = None

        if restart_data_file and restart_data4_file:
            return (restart_data_seqno, restart_data_file, restart_data4_file)
        else:
            return None


class GCellJunctionTask(RetryableTask, gc3libs.utils.Struct):
    """
    Retry execution of a `GCellJunctionApplication` if it fails.
    """
    def __init__(self, sim_no, executable=None, restart=None, **extra_args):
        self.sim_no = sim_no
        RetryableTask.__init__(
            self,
            # actual computational job
            GCellJunctionApplication(sim_no, executable, restart, **extra_args),
            # keyword arguments
            **extra_args)

    def retry(self):
        """Never resubmit a task."""
        return False

    _CHECK_LINES = 5
    def update_state(self, **extra_args):
        super(GCellJunctionTask, self).update_state(**extra_args)
        if self.execution.state == Run.State.RUNNING:
            try:
                estimated_size = gc3libs.defaults.PEEK_FILE_SIZE * self._CHECK_LINES
                with self.task.peek('stdout', offset=-estimated_size, size=estimated_size) as fd:
                    # drop first and last lines, as they may be partial
                    lines = fd.readlines()[1:-1]
                    # gc3libs.log.debug(
                    #     "Lines read from remote output: %s",
                    #     str.join(' ', [("<<%s>>" % ln) for ln in lines]))
                    for line in reversed(lines):
                        line = line.strip()
                        if line != '':
                            self.execution.info = line
                            break
            except Exception, err:
                gc3libs.log.warning(
                    "Ignored error while updating state of Task %s: %s: %s",
                    self, err.__class__.__name__, err)


## main script class

class GCellJunctionScript(SessionBasedScript):
    """
Read the specified INPUT ``.csv`` files and submit jobs according
to the content of those files.  Job progress is monitored and, when a
job is done, its ``data/`` and ``data4/`` output directories are
retrieved back into the same directory where the executable file is
(this can be overridden with the ``-o`` option).

The ``gcelljunction`` command keeps a record of jobs (submitted, executed
and pending) in a session file (set name with the ``-s`` option); at
each invocation of the command, the status of all recorded jobs is
updated, output from finished jobs is collected, and a summary table
of all known jobs is printed.  New jobs are added to the session if
new input files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; ``gcelljunction`` will delay submission of
newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            input_filename_pattern = '*.csv',
            application = GCellJunctionApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GCellJunctionTask,
            )


    def setup_options(self):
        self.add_param("-x", "--executable", metavar="PATH",
                       dest="executable", default=None,
                       help="Path to the `tricellular_junctions` executable file.")
        # change default for the memory/walltime options
        self.actions['memory_per_core'].default = 3*Memory.GB
        self.actions['wctime'].default = '60 days'


    def make_directory_path(self, pathspec, jobname):
        # XXX: Work around SessionBasedScript.process_args() that
        # apppends the string ``NAME`` to the directory path.
        # This is really ugly, but the whole `output_dir` thing needs to
        # be re-thought from the beginning...
        if pathspec.endswith('/NAME'):
            return pathspec[:-len('/NAME')]
        else:
            return pathspec


    def before_main_loop(self):
        # XXX: should this be done with `make_controller` instead?
        self._controller.retrieve_running = True
        self._controller.retrieve_overwrites = True
        self._controller.retrieve_changed_only = True


    def new_tasks(self, extra):
        # how many iterations are we already computing (per parameter set)?
        iters = defaultdict(int)
        for task in self.session:
            name, instance = task.jobname.split('#')
            iters[name] = max(iters[name], int(instance))

        for path in self.params.args:
            if path.endswith('.csv'):
                try:
                    inputfile = open(path, 'r')
                except (OSError, IOError), ex:
                    self.log.warning("Cannot open input file '%s': %s: %s",
                                     path, ex.__class__.__name__, str(ex))
                    continue
                for lineno, line in enumerate(inputfile):
                    line = line.strip()
                    # ignore blank and comment lines (those that start with '#')
                    if len(line) == 0 or line.startswith('#'):
                        continue
                    try:
                        parts = re.split("\s*[ ,]\s*", line)
                        sim_no = int(parts[0])
                        if len(parts) == 2:
                            restart = string_to_boolean(parts[1])
                        else:
                            restart = False
                    except ValueError:
                        self.log.error("Wrong format in line %d of file '%s':"
                                       " need 1 integer value (`SimNo`),"
                                       " and -optionally- a flag to restart,"
                                       " but actually got '%s'."
                                       " Ignoring input line, fix it and re-run.",
                                       lineno+1, path, line)
                        continue # with next `row`
                    # extract parameter values
                    basename = ('tricellular_junction_%d' % (sim_no,))

                    # prepare job(s) to submit
                    already = len([ task for task in self.session if task.sim_no == sim_no ])
                    kwargs = extra.copy()
                    base_output_dir = kwargs.pop('output_dir', self.params.output)
                    jobname = ('%s#%d' % (basename, already+1))
                    output_dir = os.path.join(base_output_dir, jobname)
                    if restart and os.path.isdir(output_dir):
                        self.log.debug("Looking for restart files in directory '%s' ...", output_dir)
                        restart = GCellJunctionApplication.find_restart_data(output_dir)
                    else:
                        restart = None
                    yield GCellJunctionTask(
                        sim_no,
                        executable=self.params.executable,
                        restart=restart,
                        jobname=jobname,
                        output_dir=output_dir,
                        **kwargs)

            else:
                self.log.error("Ignoring input file '%s': not a CSV file.", path)



#! /usr/bin/env python
#
# Copyright (C) 2011-2012  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
"""
"""

from __future__ import absolute_import, print_function

__author__ = 'Sergio Maffioletti <sergio.maffioletti@gc3.uzh.ch>'
# summary of user-visible changes
__changelog__ = """
  2011-06-27:
    * Defined ABCApplication and basic SessionBasedScript
"""
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == '__main__':
    import cchem_abc_uml
    cchem_abc_uml.ABC_uml_Workflow().run()


# stdlib imports
import os
import os.path
import sys

# interface to Gc3libs
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, _Script, existing_file

EXECUTABLE="/home/sergio/dev/comp.chem/xrls/abc.sh"
ABC_EXECUTABLE="/home/sergio/dev/comp.chem/xrls/abc.x"
APPPOT_IMAGE=""
APPPOT_RUN=""

ABC_UML_IMAGE="gsiftp://idgc3grid01.uzh.ch/local_repo/ABC/abc.img"
ABC_APPOTRUN="gsiftp://idgc3grid01.uzh.ch/local_repo/ABC/abc_appotrun.sh"

# & (executable = "$APPPOT_STARTUP" )
# (arguments = "--apppot" "abc.cow,abc.img")
# (jobname = "ABC_uml")
# (executables = "apppot-run")
# (gmlog=".arc")
# (join="yes")
# (stdout=".out")
# (inputFiles=("abc.img" "gsiftp://idgc3grid01.uzh.ch/local_repo/ABC/abc.img") ("nevpt2_cc-pV5Z.g3c" "./nevpt2_cc-pV5Z.g3c") ("apppot-run" "./gfit3c_abc.sh") ("dimensions" "./dimensions"))
# (outputfiles=("abc.x" "abc.nevpt2_cc-pV5Z") ("nevpt2_cc-pV5Z_log.tgz" ""))
# (runtimeenvironment = "TEST/APPPOT-0")
# (wallTime="10")
# (memory="2000")

class Gfit3C_ABC_uml_Application(Application):
    def __init__(self, abc_uml_image_file, abc_apppotrun_file, output_folder, g3c_input_file=None, dimension_file=None, surface_file=None, **extra_args):

        inputs = [(abc_uml_image_file, "abc.img"), (abc_apppotrun_file, "apppot-run")]

        if surface_file:
            inputs.append((surface_file,os.path.basename(surface_file)))
            abc_prefix = os.path.basename(surface_file)
        elif g3c_input_file and dimension_file:
            inputs.append((g3c_input_file, os.path.basename(g3c_input_file)))
            inputs.append((dimension_file,"dimensions"))
            abc_prefix = os.path.basename(g3c_input_file)
        else:
            raise gc3libs.exceptions.InvalidArgument("Missing critical argument surface file [%s], g3c_file [%s], dimension_file [%s]" % (surface_file, g3c_input_file, dimension_file))

        extra_args['tags'] = ["TEST/APPPOT-0"]

        gc3libs.Application.__init__(self,
                                     executable = "$APPPOT_STARTUP",
                                     arguments = ["--apppot", "abc.cow,abc.img"],
                                     inputs = inputs,
                                     outputs = [ ("abc.x", "abc."+abc_prefix), (os.path.basename(abc_prefix).split(".g3c")[0]+"_log.tgz", os.path.basename(abc_prefix).split(".g3c")[0]+"_log.tgz")],
                                     join = True,
                                     stdout = os.path.basename(abc_prefix).split(".g3c")[0]+".log",
                                     **extra_args
                                     )

class ABC_uml_Workflow(SessionBasedScript):
    """
    gdemo
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = Gfit3C_ABC_uml_Application,
            input_filename_pattern = '*.d',
            )


    def setup_options(self):
        # _Script.setup(self)

        self.add_param("-g", "--g3c",
                       action="store", dest="g3cfile",
                       type=existing_file, default=None,
                       help="G3C input file.")

        self.add_param("-d", "--dimensions",
                       action="store", dest="dimensions",
                       type=existing_file, default=None,
                       help="Surface file."
                       )
    #     self.add_param("-C", "--continuous", type=int, dest="wait", default=0,
    #                    metavar="INTERVAL",
    #                    help="Keep running, monitoring jobs and possibly submitting new ones or"
    #                    " fetching results every INTERVAL seconds. Exit when all jobs are finished."
    #                    )
    #     self.add_param("-w", "--wall-clock-time", dest="wctime", default=str(8), # 8 hrs
    #                    metavar="DURATION",
    #                    help="Each job will run for at most DURATION time"
    #                    " (default: %(default)s hours), after which it"
    #                    " will be killed and considered failed. DURATION can be a whole"
    #                    " number, expressing duration in hours, or a string of the form HH:MM,"
    #                    " specifying that a job can last at most HH hours and MM minutes."
    #                    )
        return

    def parse_args(self):
        self.apppot_run = self.params.args[0]
        self.output_folder = self.params.output

        # compile g3fit
        self.counter = 0



#        gc3libs.log.info("input folder [%s] output folder [%s]" % (self.input_folder,self.output_folder))

    def new_tasks(self, extra):

         extra_args = extra.copy()

         if self.params.dimensions and self.params.g3cfile and self.counter < 1:
             #     def __init__(self, abc_uml_image_file, abc_apppotrun_file, output_folder, g3c_input_file=None, dimension_file=None, surface_file=None, **extra_args):
             name = "Gfit3C_"+str(os.path.basename(self.params.g3cfile))

             yield (name, Gfit3C_ABC_uml_Application, [
                     ABC_UML_IMAGE,
                     self.apppot_run,
                     self.params.output,
                     self.params.g3cfile,
                     self.params.dimensions,
                     ], extra_args)
             self.counter = self.counter + 1


         # inputs = SessionBasedScript._search_for_input_files(self, self.params.args)

         # for path in inputs:
         #     name = "ABC_"+str(os.path.basename(path))

         #     path = os.path.abspath(path)

         #     gc3libs.log.info("Calling ABCWorkflow.next_tastk() for param [%s] ... " % path)

         #     yield (name, Gfit3C_ABC_uml_Application, [
         #             EXECUTABLE,
         #             ABC_EXECUTABLE,
         #             path,
         #             self.params.output,
         #             ], extra_args)



#! /usr/bin/env python

"""
This module implements "pollers". A "Poller" is an object that
monitors a given URL and returns `events` whenever a new object is
created inside that URL.
"""

# Copyright (C) 2017-2019,  University of Zurich. All rights reserved.
# Copyright (C) 2020,  ETH Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
from builtins import object
from abc import ABCMeta, abstractmethod
from collections import defaultdict
import os
import stat
from warnings import warn

# GC3Pie imports
import gc3libs
from gc3libs.url import Url
from future.utils import with_metaclass


# Some pollers depend on the presence of specific Python modules;
# delay ImportError until actually used but raise a warning.
class _Unavailable(object):
    def __init__(self, missing):
        self.__missing = missing
        warn("Trying to initialize `{module}` which is not available."
             " A placeholder object will be used instead, but it will raise"
             " `ImportError` later if there is any actual attempt at using it."
             .format(module=self.__missing),
             ImportWarning)
    def __getattr__(self, name):
        raise ImportError(
            "Trying to actually use module `{module}`"
            " which could not be imported. Aborting."
            .format(module=self.__missing))

try:
    import inotify_simple as inotify
except:
    inotify = _Unavailable('inotify_simple')

try:
    import swiftclient
except ImportError:
    swiftclient = _Unavailable('swiftclient')


##
#
# Registration support
#
##

_AVAILABLE_POLLERS = defaultdict(list)

def make_poller(url, **extra):
    """
    Factory method that returns the registered poller for the specified
    :py:mod:`gc3libs.url.Url`.
    """

    url = Url(url)
    try:
        registered = _AVAILABLE_POLLERS[url.scheme]
    except KeyError:
        raise ValueError(
            "No poller associated with scheme `{0}`"
            .format(url.scheme))
    for cls in registered:
        try:
            poller = cls(url, **extra)
            gc3libs.log.debug("Using class %s to poll URL %s", cls, url)
            return poller
        except Exception as err:
            gc3libs.log.debug(
                "Could not create poller for scheme `%s` with class %s: %s",
                url.scheme, cls, err)
    raise ValueError(
        "No registered class could be instanciated to poll URL {0}"
        .format(url))


def register_poller(scheme):
    def register_class(cls):
        _AVAILABLE_POLLERS[scheme].append(cls)
        return cls
    return register_class


class Poller(with_metaclass(ABCMeta, object)):
    """
    Abstract class for an URL Poller.

    A :py:class:`Poller` is a class that tracks new events on a
    specific :py:class:`Url`. When calling the :py:meth:`get_events()`
    it will return a list of tuples (Url, mask) containing the events
    occurred for each one of the underlying URLs.
    """

    __slots__ = ['url']

    def __init__(self, url, **kw):
        self.url = Url(url)

    @abstractmethod
    def get_new_events(self):
        """
        Iterate over events that happened since last call to this method.

        Returns a list of tuples *(subject, event)*.

        A *subject* is a unique identifier for a watched "thing": the
        exact form and type depends on the actual concrete class;
        pollers that watch the filesystem or HTTP-accessible resources
        will use a URL (:class:`gc3libs.url.Url`) as a *subject*
        instance, but e.g. pollers that watch a database table might
        use a row ID instead.

        The associated *event* is one or more of the following
        strings:

        - ``created``: the subject has been created since the last
          call to ``get_new_events()``;

        - ``modified``: the subject has changed since the last call to
          ``get_new_events()``;

        - ``deleted``: the subject has been deleted since the last
          call to ``get_new_events()``;

        Depending on the concrete poller class, some events might
        never occur, or cannot be detected.  Most notably, only
        filesystem-watching pollers might be able to generate
        meaningful ``modified`` events.
        """
        raise NotImplementedError(
            "Abstract method `Poller.get_events()` called "
            " - this should have been defined in a derived class.")


##
#
# Filesystem polling
#
##

@register_poller('file')
class INotifyPoller(Poller):
    """
    Use Linux' inofity to track new events on the specified filesystem location.

    :params recurse:
      When ``True``, automatically track events in any
      (already-existing or newly-created) subdirectory.

    This poller is used by default when the `inotify_simple` Python
    package is available and the URL has a `file` schema.

    .. warning::

        On Linux, the maximum number of inotify descriptors that
        a user can open is limited by the kernel parameters:

        * ``fs.inotify.max_user_instances``
        * ``fs.inotify.max_user_watches``
        * ``fs.inotify.max_queued_events``

        See also the `inotify(7) manpage <http://linux.die.net/man/7/inotify>`_
    """

    __slots__ = [
        '_ifd',
        '_recurse',
        '_wd',
    ]

    def __init__(self, url, recurse=False, **kw):
        # normalize filesystem paths by removing trailing slashes;
        # this is necessary for the check at the beginning of
        # `self.watch()` to work correctly when path == self.url.path
        try:
            path = os.path.abspath(url.path)
        except AttributeError:
            path = os.path.abspath(url)
        super(INotifyPoller, self).__init__(path, **kw)

        self._recurse = recurse
        self._ifd = inotify.INotify()
        self._wd = {}

        # Ensure inbox directory exists
        if not os.path.exists(self.url.path):
            gc3libs.log.info(
                "Inbox directory `%s` does not exist,"
                " creating it ...", self.url.path)
            os.makedirs(self.url.path)

        self.watch(self.url.path)
        if self._recurse:
            for dirpath, dirnames, filenames in os.walk(self.url.path):
                self.watch(dirpath)
                for name in filenames:
                    path = os.path.join(self.url.path, dirpath, name)
                    self.watch(path)

    @property
    def recurse(self):
        """
        Whether the poller is watching the entire directory tree pointed
        to by ``self.url``, or only the directory at its top level.
        """
        return self._recurse

    def watch(self, path):
        path = os.path.abspath(path)
        # FIXME: this is not enough, as we might be adding a watch to
        # a file located in a subdirectory of `self.url` without *recurse*
        assert path.startswith(self.url.path), (
            "Request to watch path `%s`"
            " which does not lie in directory tree rooted at `%s`",
            path, self.url.path)
        wd = self._ifd.add_watch(path,
                                 # only watch for inotify(7) events that
                                 # translate to something we can report
                                 (0
                                  |inotify.flags.CLOSE_WRITE
                                  |inotify.flags.CREATE
                                  |inotify.flags.DELETE
                                  |inotify.flags.EXCL_UNLINK
                                  |inotify.flags.MODIFY
                                  |inotify.flags.OPEN
                                 ))
        self._wd[wd] = path
        gc3libs.log.debug(
            "Added watch for path `%s` as watch descriptor %d ", path, wd)

    def get_new_events(self):
        new_events = []
        ievents = self._ifd.read(0)
        cumulative = defaultdict(int)
        for ievent in ievents:
            try:
                path = os.path.join(self._wd[ievent.wd], ievent.name)
            except KeyError:
                raise AssertionError(
                    "Received event {0} for unknown watch descriptor {1}"
                    .format(ievent, ievent.wd))
            # if `name` is empty, it's the same directory
            cumulative[path] |= ievent.mask
            accumulated = cumulative[path]
            # we want to dispatch a single `created` or `modified`
            # event, so check once the file is closed what past events
            # have been recorded
            if (ievent.mask & inotify.flags.CLOSE_WRITE):
                if (accumulated & inotify.flags.CREATE):
                    new_events.append(
                        self.__make_event(path, 'created'))
                elif (accumulated & inotify.flags.MODIFY):
                    new_events.append(
                        self.__make_event(path, 'modified'))
            if (ievent.mask & inotify.flags.DELETE):
                new_events.append(
                    self.__make_event(path, 'deleted'))
            if (self._recurse
                and ievent.mask & inotify.flags.ISDIR
                and ievent.mask & inotify.flags.CREATE):
                # A new directory has been created. Add a watch
                # for it too and for all its subdirectories. Also,
                # we need to trigger new events for any file
                # created in it.
                for (rootdir, dirnames, filenames) in os.walk(path):
                    for dirname in dirnames:
                        self.watch(os.path.join(rootdir, dirname))
                    for filename in filenames:
                        # report creation event
                        new_events.append(
                            (Url(os.path.join(rootdir, filename)),
                             'created'))
        return new_events

    def __make_event(self, relpath, event):
        if relpath:
            url = Url(os.path.join(self.url.path, relpath))
        else:
            url = self.url
        return (url, event)


@register_poller('file')
class FilePoller(Poller):
    """
    Track events on the filesystem using Python's standard `os` module.

    :params recurse:
      When ``True``, automatically track events in any
      (already-existing or newly-created) subdirectory.

    .. warning::

      In order to issue `'modified'` events, this class relies on
      checking an inode's `st_mtime` field, which only provides
      1-second resolution.  Modification events that happen too close
      will not be told apart as distinct; in particular, modifying a
      file less than 1s after creating it will not be detected.

    This implementation is only used to track Url with `file` schema
    whenever the :py:mod:`inotify_simple` module is not available.
    """

    __slots__ = [
        '_recurse',
        '_watched'
    ]

    def __init__(self, url, recurse=False, **kw):
        super(FilePoller, self).__init__(url, **kw)

        root = self.url.path
        if not os.path.exists(root):
            gc3libs.log.warning(
                "Inbox root directory `%s` does not exist,"
                " creating it ...", root)
            os.makedirs(root)
        self._recurse = recurse
        self._watched = {}

        self.watch(root)

    @property
    def recurse(self):
        """
        Whether the poller is watching the entire directory tree pointed
        to by ``self.url``, or only the directory at its top level.
        """
        return self._recurse

    def watch(self, path):
        info = os.stat(path)
        self._watched[path] = info.st_mtime
        if stat.S_ISDIR(info.st_mode) and self._recurse:
            for entry in os.listdir(path):
                if entry in ['.', '..']:
                    continue
                child_path = os.path.join(path, entry)
                self.watch(child_path)


    def get_new_events(self):
        new_events = []
        # take a copy of the list of watched files, since
        # `self._get_events_*` may modify the `self._watched` dict
        for path in list(self._watched.keys()):
            if os.path.isdir(path):
                new_events += self._get_events_dir(path)
            else:
                new_events += self._get_events_file(path)
        return new_events


    def _get_events_file(self, path):
        if not os.path.exists(path):
            self._watched.pop(path)
            return [(Url(path), 'deleted')]

        event = None
        mtime = os.stat(path).st_mtime
        if mtime > self._watched[path]:
            event = 'modified'
        self._watched[path] = mtime

        if event:
            return [(Url(path), event)]
        else:
            return []

    def _get_events_dir(self, dirpath):
        new_events = []

        # check for new files
        contents = set(
            os.path.join(dirpath, entry)
            for entry in os.listdir(dirpath)
            if entry not in ['.', '..']
        )
        for path in contents:
            if path not in self._watched:
                new_events.append((Url(path), 'created'))
                # if path is a directory, add it to watch list only if
                # poller was created with `recurse=True`
                if (not os.path.isdir(path)) or self._recurse:
                    self.watch(path)

        return new_events


class SwiftPoller(Poller):
    """
    Periodically check a SWIFT bucket and trigger
    events when new objects are created.

    Right now, a valid URL can be one of the following form:

    - If the keystone endpoint is reachable via HTTP, either one of:

      * swift://<user>+<tenant>:<password>@<keystone-url>?container
      * swt://<user>+<tenant>:<password>@<keystone-url>?container

    - If the keystone endpoint is reachable via HTTPS, either one of:

      * swifts://<user>+<tenant>:<password>@<keystone-url>?container
      * swts://<user>+<tenant>:<password>@<keystone-url>?container

    We assume that keystone auth version 2 is used.
    """
    def __init__(self, url, **kw):
        super(SwiftPoller, self).__init__(url, **kw)

        try:
            self.username, self.project_name = self.url.username.split('+')
        except ValueError:
            raise gc3libs.exceptions.InvalidValue(
                "Missing project/tenant name in SWIFT URL '{0}'"
                .format(self.url))
        self.password = self.url.password
        self.container = self.url.query

        if not self.container:
            raise gc3libs.exceptions.InvalidValue(
                "Missing bucket name in SWIFT URL '{0}'"
                .format(self.url))
        # FIXME: also check for hostname and password?

        if url.scheme in ['swifts', 'swts']:
            auth_url = 'https://%s' % self.url.hostname
        else:
            auth_url = 'http://%s' % self.url.hostname
        if self.url.port:
            auth_url += ":%d" % self.url.port
        if self.url.path:
            auth_url += self.url.path
        self.auth_url = auth_url

        self.conn = swiftclient.Connection(
            authurl=self.auth_url,
            user=self.username,
            key=self.password,
            os_options = {
                "auth_url": self.auth_url,
                "project_name": self.project_name,
                "username": self.username,
                "password": self.password},
            auth_version='2')

        # List containers
        _, containers = self.conn.get_account()
        gc3libs.log.debug(
            "Successfully connected to SWIFT storage '%s':"
            " %d containers found", self.conn.url, len(containers))
        if self.container not in [a.get('name') for a in containers]:
            gc3libs.log.warning(
                "Container %s not found at SWIFT URL '%s'",
                self.container, self.url)

        _, objects = self.conn.get_container(self.container)
        self._known_objects = {}
        for obj in objects:
            url = Url(str(self.url) + '&name=%s' % obj['name'])
            self._known_objects[url] = obj

    def get_new_events(self):
        # List objects in container
        _, objects = self.conn.get_container(self.container)
        newevents = []

        objurls = []
        for obj in objects:
            url = Url('{baseurl}&name={objname}'
                      .format(baseurl=self.url,
                              objname=obj['name']))
            objurls.append(url)
            if url not in self._known_objects:
                self._known_objects[url] = obj
                newevents.append((url, 'created'))
        for url in list(self._known_objects):
            if url not in objurls:
                newevents.append((url, 'deleted'))
                self._known_objects.pop(url)
        return newevents


# register for multiple schemes
register_poller('swift')(SwiftPoller)
register_poller('swifts')(SwiftPoller)
register_poller('swt')(SwiftPoller)
register_poller('swts')(SwiftPoller)


# main: run tests
if "__main__" == __name__:
    import doctest
    doctest.testmod(name="poller",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   gparseurl.py -- Front-end script for submitting multiple `GC-GPS` R-baseed jobs.
#
#   Copyright (C) 2011, 2012  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `R` jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gparseurl.py --help`` for program usage instructions.

Input parameters consists of:
:param str URL_file: Path to the file containing the list of the URL to be processed
   example:
https://archive.org/download/archiveteam-twitter-stream-2014-06/archiveteam-twitter-stream-2014-06.tar

Option:
:param str parser_script: Path to the parser script. Default: None.

"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2015-10-24:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gparseurl
    gparseurl.GParseURLScript().run()

import os
import sys
import time
import tempfile

import shutil

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask


## custom application class
class GParseURLApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'gparseurl'

    def __init__(self, url_to_read, **extra_args):


        inputs = dict()
        outputs = dict()

        if 'master_script' in extra_args:
            inputs[extra_args['master_script']] = "./master.py"

        command = "python ./master.py %s" % url_to_read

        outputs['result.json'] = "%s.json" % extra_args['jobname']

        Application.__init__(
            self,
            arguments = command,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gpaseurl.log',
            join=True,
            **extra_args)


class GParseURLScript(SessionBasedScript):
    """
Read URL_list file containing the list of URLs to be processed.
For each URL - or a chunk of them - create a GParseURLApplication
to run the parser script in a parallel manner. Job progress is
monitored and, when a job is done, its output files are retrieved back
into the simulation directory itself.

The ``gparseurl`` command keeps a record of jobs (submitted, executed
and pending) in a session file (set name with the ``-s`` option); at
each invocation of the command, the status of all recorded jobs is
updated, output from finished jobs is collected, and a summary table
of all known jobs is printed.  New jobs are added to the session if
new input files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; ``gparseurl`` will delay submission of
newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GParseURLApplication,
            stats_only_for = GParseURLApplication,
            )

    def setup_options(self):
        self.add_param("-Z", "--master", metavar="PATH",
                       dest="master_script", default=None,
                       help="Path to the parser script. Default: None.")

    def setup_args(self):

        self.add_param('url_file', type=str,
                       help="Command file full path name")

    def parse_args(self):
        """
        Check presence of input folder (should contains R scripts).
        path to url_file should also be valid.
        """

        if not os.path.exists(self.params.url_file):
            raise gc3libs.exceptions.InvalidUsage(
                "gparseurl command file '%s' does not exist;"
                % self.params.url_file)
        gc3libs.utils.check_file_access(self.params.url_file, os.R_OK,
                                gc3libs.exceptions.InvalidUsage)

        if self.params.master_script and \
        not os.path.exists(self.params.master_script):
            raise gc3libs.exceptions.InvalidUsage(
                "Input folder '%s' does not exists"
                % self.params.master_script)

    def new_tasks(self, extra):
        """
        Read content of 'url_file'
        For each command line, generate a new GcgpsTask
        """

        tasks = []

        for url_to_read in open(self.params.url_file):
            jobname = os.path.splitext(os.path.basename(url_to_read))[0]
            extra_args = extra.copy()
            extra_args['jobname'] = jobname
            extra_args['output_dir'] = self.params.output

            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', os.path.join('.computation',jobname))
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', os.path.join('.computation',jobname))
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', os.path.join('.computation',jobname))
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', os.path.join('.computation',jobname))

            if self.params.master_script:
                extra_args['master_script'] = self.params.master_script

            self.log.debug("Creating Task for %s" % jobname)

            tasks.append(GParseURLApplication(
                url_to_read,
                **extra_args))

        return tasks



#!/bin/env python
#
# gbenchmark_wrapper_allinone.py -- wrapper script for executing igraph R script
#
# Authors: Sergio Maffioletti <sergio.maffioletti@uzh.ch>
#
# Copyright (C) 2015  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from __future__ import absolute_import, print_function, unicode_literals
import sys
import os
import subprocess

DEFAULT_RESULT_FOLDER = "./results"

def usage():
    print("""
    Usage:
    gbenchmark_allinone.py <benchmark_file> <network files folder>
    """)

def RunBenchmarks(benchmark_file, network_files_folder):

    # create result fodler
    os.makedirs(DEFAULT_RESULT_FOLDER)
    returncode = 0

    # scan through network_files_folder and run benchmark for each network file
    for network in os.listdir(network_files_folder):
        os.makedirs(os.path.join(DEFAULT_RESULT_FOLDER,network))
        command="Rscript --vanilla %s %s %s" % (benchmark_file,
                                                os.path.join(network_files_folder,
                                                             network),
                                                os.path.join(DEFAULT_RESULT_FOLDER,
                                                             network))
        proc = subprocess.Popen(
            [command],
            shell=True,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE)

        print("Running command %s" % command)
        (stdout, stderr) = proc.communicate()

        if proc.returncode != 0:
            print("Execution failed with exit code: %d" % proc.returncode)

            print(stdout)
            print(stderr)

            returncode = proc.returncode

if __name__ == '__main__':
    if len(sys.argv[1:]) != 2:
           usage()
           sys.exit(1)
    sys.exit(RunBenchmarks(sys.argv[1],sys.argv[2]))



#! /usr/bin/env python
#
#   gpredict_PopContCC.py -- Front-end script for running Matlab function
#   function over a large parameter range.
#
#   Copyright (C) 2016  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for running Matlab function
#   function over a large parameter range.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gpredict_PopContCC.py --help`` for program usage
instructions.

"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2016-08-17:
  * Initial version
  2016-08-19:
  * add '-f <function name>' option
  2016-09-30:
  * add extra option 'bundle' and option value
    to the matlab function to be called
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == "__main__":
    import gpredict_PopContCC
    gpredict_PopContCC.Gpredict_PopContCCScript().run()

import os
import sys
import time
import tempfile

import tarfile
import shutil

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file, positive_int, existing_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_MATLAB_FUNCTION="predict_PopContCC"
DEFAULT_REMOTE_OUTPUT_FOLDER = "./results"
TARFILE="source.tgz"
TEMP_FOLDER="/var/tmp"
DEFAULT_REPETITIONS = 100
DEFAULT_BUNDLE = 10
MATLAB_COMMAND="matlab -nodesktop -nodisplay -nosplash -r \'{mfunct} {MatPredictor} " \
    "{VecResponse} {numberOfSamples} {numberOfTrees} {bundles} {results} ;quit()'"

## utility funtions
def _get_iterations(repetitions,bundle):
    """
    Yield iterator number and bundle size.
    Last bundle value may be shorter than 'bundle'
    """
    total_iterations = repetitions/bundle
    last_iteration = repetitions%bundle

    for iteration in range(1,total_iterations+1):
        yield (iteration,bundle)

    if repetitions%bundle > 0:
        yield (total_iterations+1,repetitions%bundle)


def _scanandtar(dir_to_scan, temp_folder=TEMP_FOLDER):
    try:
        gc3libs.log.debug("Compressing input folder '%s'" % dir_to_scan)
        cwd = os.getcwd()
        os.chdir(dir_to_scan)

        if not os.path.isdir(temp_folder):
            os.mkdir(temp_folder)

        with tarfile.open(os.path.join(temp_folder,TARFILE), "w:gz") as tar:

            tar.add(dir_to_scan, arcname=".")
            os.chdir(cwd)

            gc3libs.log.info("Created tar file '%s'" % TARFILE)
            return tar.name

    except Exception, x:
        gc3libs.log.error("Failed creating input archive '%s': %s %s",
                          os.path.join(dir_to_scan,),
                          type(x),x.message)
        raise


## custom application class

class Gpredict_PopContCCApplication(Application):
    """
    Custom class to wrap the execution of the matlab function.
    """
    application_name = 'gpredictpopcontcc'

    def __init__(self, Mfunct, MatPredictor_file, VecResponse_file,
                 numberOfSamples, numberOfTrees, bundles, iteration, **extra_args):

        executables = []
        inputs = dict()
        outputs = dict()

        self.iteration = iteration
        self.results = extra_args['session_output_dir']

        inputs[Mfunct] = os.path.basename(Mfunct)
        function_name = inputs[Mfunct].split('.')[0]
        inputs[MatPredictor_file] = os.path.basename(MatPredictor_file)
        inputs[VecResponse_file] = os.path.basename(VecResponse_file)

        arguments = MATLAB_COMMAND.format(mfunct=function_name,
                                          MatPredictor=inputs[MatPredictor_file],
                                          VecResponse=inputs[VecResponse_file],
                                          numberOfSamples=numberOfSamples,
                                          numberOfTrees=numberOfTrees,
                                          bundles=bundles,
                                          results=DEFAULT_REMOTE_OUTPUT_FOLDER)

        # Set output
        outputs[DEFAULT_REMOTE_OUTPUT_FOLDER] = DEFAULT_REMOTE_OUTPUT_FOLDER

        gc3libs.log.debug("Creating application for executing: %s",
                          arguments)

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gpredict_PopContCC.log',
            join=True,
            executables = executables,
            **extra_args)

    def terminated(self):
        folder_files = os.path.join(self.output_dir, DEFAULT_REMOTE_OUTPUT_FOLDER)
        for f in os.listdir(folder_files):
            os.rename(os.path.join(folder_files ,f),
                      os.path.join(self.results, "%d_%s" % (self.iteration,f)))

class Gpredict_PopContCCScript(SessionBasedScript):
    """
    Takes a matlab function file as first argument, then MatPredictor .mat file
    then a VecResponse .mat file, then numberOfSamples and numberOfTrees.
    Additionally takes the number of repetitions as option.

    For each repetitions, the script executed the function specified in the matlab
    function.
    expected function signature:
    predict_PopContCC(matPredictor,vecResponse,numberOfSamples,numberOfTrees,resultFolder)

    The ``gpredict_PopContCC`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gpredict_PopContCC`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gpredict_PopContCC``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = Gpredict_PopContCCApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = Gpredict_PopContCCApplication,
            )

    def setup_args(self):
        self.add_param('Mfunct',
                       type=existing_file,
                       help="Path to the gpredict_PopContCC file.")

        self.add_param('MatPredictor',
                       type=existing_file,
                       help="Path to the MatPredictor file.")

        self.add_param('VecResponse',
                       type=existing_file,
                       help="Path to the VecResponse file.")


        self.add_param('numberOfSamples',
                       type=positive_int,
                       help="Number of samples.")

        self.add_param('numberOfTrees',
                       type=positive_int,
                       help="Number of trees.")

    def setup_options(self):
        self.add_param("-R", "--repetitions",
                       metavar="[INT]",
                       type=positive_int,
                       dest="repetitions",
                       default=DEFAULT_REPETITIONS,
                       help="Number of repetitions. "
                       " Repeat Matlab execution. "
                       " Default: repeat '%(default)s' times.")

        self.add_param("-B", "--bundle",
                       metavar="[INT]",
                       type=positive_int,
                       dest="bundle",
                       default=DEFAULT_BUNDLE,
                       help="Group execution of repetitions in bundles. "
                       " Total executions: repetitions / bundle. "
                       " Default: bundle in group of '%(default)s' repetitions.")


    def new_tasks(self, extra):
        """
        Read content of 'command_file'
        For each command line, generate a new Application
        """
        tasks = []

        for iteration,bundle_size in _get_iterations(self.params.repetitions,self.params.bundle):
            jobname = "gpredict_PopContCC-%d" % (iteration)

            extra_args = extra.copy()
            extra_args['jobname'] = jobname
            extra_args['output_dir'] = self.params.output.replace('NAME', jobname)
            extra_args['session_output_dir'] = os.path.dirname(self.params.output)

            self.log.debug("Creating Application for iteration : %d" % iteration)

            tasks.append(Gpredict_PopContCCApplication(
                os.path.abspath(self.params.Mfunct),
                os.path.abspath(self.params.MatPredictor),
                os.path.abspath(self.params.VecResponse),
                self.params.numberOfSamples,
                self.params.numberOfTrees,
                bundle_size,
                iteration,
                **extra_args))

        return tasks



#! /usr/bin/env python
#
"""
Tests for the cmdline module
"""
# Copyright (C) 2012-2016, 2018, 2019 S3IT, Zentrale Informatik, University of
# Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import absolute_import, print_function, unicode_literals
from builtins import range
from builtins import object
import json
import os
from os.path import abspath, dirname, exists, isdir, isfile, join
import re
import shutil
import signal
import subprocess
import sys
import tempfile
import time

import pytest

import gc3libs
import gc3libs.session
from gc3libs.utils import read_contents, write_contents, to_str


class _TestsCommon(object):
    """
    Set up environment for testing session-based commands.
    """

    @pytest.fixture(autouse=True)
    def setUp(self):
        self.scriptdir = join(dirname(abspath(__file__)), 'scripts')
        self.client_py = join(self.scriptdir, 'simpleclient.py')
        self.daemon_py = join(self.scriptdir, 'simpledaemon.py')
        self.script_py = join(self.scriptdir, 'simplescript.py')

        self.basedir = tempfile.mkdtemp(
            prefix=(__name__ + '.'),
            suffix='.tmp.d')
        orig_wd = os.getcwd()
        os.chdir(self.basedir)

        self.cfgfile = join(self.basedir, 'config.ini')
        write_contents(self.cfgfile,
                       """
[resource/localhost]
enabled = true
auth = none
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 8GB
max_walltime = 8h
max_cores = 2
architecture = x86_64
override = False
resourcedir = {basedir}/resource.d
                       """.format(basedir=self.basedir))

        try:
            yield
        finally:
            os.chdir(orig_wd)
            shutil.rmtree(self.basedir, ignore_errors=True)

    def run(self, cmd):
        """
        Run `cmd` and return exit code and STDOUT+STDERR.
        """
        # FIXME: cannot use `subprocess.check_call` as it's not
        # available on Py 2.6
        proc = subprocess.Popen(
            # same Python that's running tests
            [sys.executable] + cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        stdout = to_str(proc.communicate()[0], 'terminal')
        rc = proc.returncode
        gc3libs.log.debug(
            "Log of command `%s`:\n```\n%s\n```",
            ' '.join(cmd), stdout)
        assert rc == 0, (
            "Command `{0}` exited with non-zero exitcode {1}"
            .format(' '.join(cmd), rc))
        return (rc, stdout)

    def check_with_timeout(self, condition, timeout, errmsg):
        """
        Repeatedly test `condition` until it either returns ``True``
        or `timeout` seconds have passed.
        """
        for _ in range(timeout):
            if condition():
                return
            time.sleep(1)
        else:
            assert False, (
                (errmsg + " (within {timeout} seconds)")
                .format(timeout=timeout))


class TestSessionBasedScript(_TestsCommon):
    """
    Test suite for session-based (interactive) scripts.
    """

    def test_session_based_script(self):
        """
        Test a very simple script based on `SessionBasedScript`:class:

        The script is found in ``gc3libs/tests/scripts/simplescript.py``
        """

        rc, stdout = self.run([
            self.script_py,
            '-C', '1',
            '-s', 'session',
            '--config-files', self.cfgfile,
            '-r', 'localhost',
        ])

        assert re.match(
                r'.*TERMINATED\s+3/3\s+\(100.0+%\).*',
                stdout,
                re.S)

        assert isdir(join(self.basedir, 'SimpleScript.out.d'))
        assert isfile(join(self.basedir,
                           'SimpleScript.out.d',
                           'SimpleScript.stdout'))

        assert isdir(join(self.basedir, 'SimpleScript.out2.d'))
        assert isfile(join(self.basedir,
                           'SimpleScript.out2.d',
                           'SimpleScript.stdout'))

        session_dir = join(self.basedir, 'session')
        assert isdir(session_dir)

        assert isfile(join(session_dir,
                           gc3libs.session.Session.INDEX_FILENAME,))

        assert isfile(join(session_dir,
                           gc3libs.session.Session.STORE_URL_FILENAME,))


# daemon-related tests fail on Py2.6 because (apparently) python-daemon
# does not correctly create PID files on that platform; disable
# the tests altogether, as we'll have to drop Py2.6 support soon...
@pytest.mark.skipif(sys.version_info < (2,7),
                    reason="python-daemon requires python 2.7+")
class TestSessionBasedDaemon(_TestsCommon):
    """
    Test suite for session-based daemons.
    """

    @pytest.fixture
    def daemon(self):
        inbox_dir = join(self.basedir, 'inbox')
        session_dir = join(self.basedir, 'session')

        subprocess.call([
            sys.executable,  # same Python that's running the tests
            join(self.scriptdir, 'simpledaemon.py',),
            '--config-files', self.cfgfile,
            '-C', '1',
            '-s', 'session',
            '--working-dir', self.basedir,
            '-r', 'localhost',
            '-vvv',
            inbox_dir,
        ])

        pidfile = join(self.basedir, 'session', 'simpledaemon.pid')
        self.check_with_timeout(
            condition=lambda: exists(pidfile),
            timeout=15,
            errmsg="PID file not created")
        pid = int(read_contents(pidfile))

        try:
            # transfer control back to test code
            yield pid, session_dir, inbox_dir
        finally:
            try:
                # kill the daemon and clean up
                os.kill(pid, signal.SIGTERM)
                # give it time to shut down
                time.sleep(3)
            except OSError as ex:
                if ex.errno == 3:  # "No such process"
                    # daemon has already exited, ignore
                    pass
                else:
                    raise


    def test_session_based_daemon_command_hello(self, daemon):
        pid, session_dir, _ = daemon

        _, stdout = self.run([self.client_py, session_dir, 'hello'])
        assert stdout.startswith("HELLO")


    def test_session_based_daemon_new_tasks(self, daemon, max_wait=10):
        pid, session_dir, _ = daemon

        # wait up to max_wait seconds for task to complete
        def check_for_terminated():
            _, stdout = self.run([self.client_py, session_dir, 'stats', 'json'])
            stats = json.loads(stdout)
            return stats.get("TERMINATED", 0) > 0
        self.check_with_timeout(
            condition=check_for_terminated,
            timeout=max_wait,
            errmsg="Daemon didn't complete initial task"
        )

        # check that output directory is there
        output_dir = join(self.basedir, 'EchoApp')
        self.check_with_timeout(
            condition=lambda: isdir(output_dir),
            timeout=max_wait,
            errmsg="Daemon didn't write output directory"
        )


    def test_session_based_daemon_inbox(self, daemon, max_wait=10):
        _, _, inboxdir = daemon

        # Wait until the daemon is up and running.
        # It will create the directory, so let's wait until then.
        self.check_with_timeout(
            condition=lambda: isdir(inboxdir),
            timeout=max_wait,
            errmsg="Daemon didn't create inbox dir"
        )

        # create marker file
        marker_file = join(inboxdir, 'foo')
        write_contents(marker_file, 'whatever')

        # wait up to max_wait seconds for task to complete
        output_dir = join(self.basedir, 'LSApp.foo')
        self.check_with_timeout(
            condition=lambda: isdir(output_dir),
            timeout=max_wait,
            errmsg="Daemon didn't process incoming file"
        )


    def test_session_based_daemon_command_help(self, daemon):
        pid, session_dir, _ = daemon

        rc, stdout = self.run([self.client_py, session_dir, 'help'])
        assert ("The following daemon commands are available:" in stdout)


    def test_session_based_daemon_command_quit(self, daemon, max_wait=60):
        pid, session_dir, _ = daemon

        rc, stdout = self.run([self.client_py, session_dir, 'quit'])

        # wait up to max_wait seconds for daemon to shut down and
        # remove PID file
        pidfile = join(session_dir, 'simpledaemon.pid')
        self.check_with_timeout(
            condition=lambda: not exists(pidfile),
            timeout=max_wait,
            errmsg="Daemon didn't shut down"
        )


    def test_session_based_daemon_reload_session(self, max_wait=60):
        inbox_dir = join(self.basedir, 'inbox')
        session_dir = join(self.basedir, 'session')

        # start daemon
        self.run([
            self.daemon_py,
            '--config-files', self.cfgfile,
            '-C', '1',
            '-s', 'session',
            '--working-dir', self.basedir,
            '-r', 'localhost',
            '-vvv',
            inbox_dir,
        ])

        pidfile = join(self.basedir, 'session', 'simpledaemon.pid')
        self.check_with_timeout(
            condition=lambda: exists(pidfile),
            timeout=15,
            errmsg="PID file not created")
        pid = int(read_contents(pidfile))

        # get IDs of tasks in session -- retry until the server is ready to respond
        for _ in range(15):
            try:
                _, stdout = self.run([self.client_py, session_dir, 'list', 'json'])
                task_ids = json.loads(stdout)
            except ValueError:
                time.sleep(1)
        assert len(task_ids) > 0

        # wait up to max_wait seconds for task to complete
        output_dir = join(self.basedir, 'EchoApp')
        self.check_with_timeout(
            condition=lambda: isdir(output_dir),
            timeout=max_wait,
            errmsg="Daemon didn't complete initial task"
        )

        # stop daemon
        self.run([self.client_py, session_dir, 'quit'])

        # wait up to max_wait seconds for daemon to shut down and
        # remove PID file
        pidfile = join(session_dir, 'simpledaemon.pid')
        self.check_with_timeout(
            condition=lambda: not exists(pidfile),
            timeout=max_wait,
            errmsg="Daemon didn't shut down"
        )

        # restart daemon
        self.run([
            self.daemon_py,
            '--config-files', self.cfgfile,
            '-C', '1',
            '-s', 'session',
            '--working-dir', self.basedir,
            '-r', 'localhost',
            '-vvv',
            inbox_dir,
        ])

        # it may take a few seconds before the daemon is ready to serve requests,
        # so try and possibly re-try...
        for _ in range(15):
            try:
                _, stdout = self.run([self.client_py, session_dir, 'list', 'json'])
                task_ids2 = json.loads(stdout)
            except ValueError:
                time.sleep(1)

        # check that task IDs are the same
        assert set(task_ids) == set(task_ids2)

        # check that one TERMINATED task is there
        _, stdout = self.run([self.client_py, session_dir, 'stats', 'json'])
        stats = json.loads(stdout)
        assert stats["TERMINATED"] == len(task_ids)


## main: run tests

if "__main__" == __name__:
    import pytest
    pytest.main(["-v", __file__])



#! /usr/bin/env python
#
"""
Test for classes and functions in the `utils` module.
"""
# Copyright (C) 2012, 2013, 2019,  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import next
from builtins import range
from builtins import object
__docformat__ = 'reStructuredText'




# 3rd party imports
import mock
import pytest

# GC3Pie imports
import gc3libs.exceptions
from gc3libs.quantity import Duration, Memory
import gc3libs.utils


# test definitions

def test_get_linux_memcg_limit_no_proc_self_cgroup():
    with mock.patch('gc3libs.utils.open') as mo:
        mo.side_effect = OSError(2, 'No such file or directory', '/proc/self/cgroup')
        limit = gc3libs.utils.get_linux_memcg_limit()
        mo.assert_called_once_with('/proc/self/cgroup', 'r')
        assert limit is None


def test_get_linux_memcg_limit_no_memcg():
    mo = mock.mock_open(read_data='*** no memcg ***')
    with mock.patch('gc3libs.utils.open', mo):
        limit = gc3libs.utils.get_linux_memcg_limit()
        mo.assert_called_once_with('/proc/self/cgroup', 'r')
        assert limit is None


def test_get_linux_memcg_limit_with_memcg_limit():
    def fake_open(path, mode):
        from io import StringIO
        from contextlib import closing
        if path == '/proc/self/cgroup':
            return closing(StringIO('2:memory:/test'))
        elif path == '/sys/fs/cgroup/memory/test/memory.soft_limit_in_bytes':
            raise OSError(2, 'No such file or directory',
                          '/sys/fs/cgroup/memory/test/memory.soft_limit_in_bytes')
        elif path == '/sys/fs/cgroup/memory/test/memory.limit_in_bytes':
            return closing(StringIO('42'))
        else:
            raise AssertionError(
                "Unexpected call to open({0!r}, {1!r})"
                .format(path, mode))
    with mock.patch('gc3libs.utils.open') as mo:
        mo.side_effect = fake_open
        limit = gc3libs.utils.get_linux_memcg_limit()
        mo.assert_has_calls([
            mock.call('/proc/self/cgroup', 'r'),
            mock.call('/sys/fs/cgroup/memory/test/memory.soft_limit_in_bytes', 'r'),
            mock.call('/sys/fs/cgroup/memory/test/memory.limit_in_bytes', 'r'),
        ])
        assert limit == Memory(42, Memory.B)


def test_parse_linux_proc_limits():
    # snapshot of `/proc/self/limits` taken on Linux 4.4.0-87-generic (Ubuntu 16.04)
    data = """\
Limit                     Soft Limit           Hard Limit           Units
Max cpu time              unlimited            unlimited            seconds
Max file size             unlimited            unlimited            bytes
Max data size             unlimited            unlimited            bytes
Max stack size            8388608              unlimited            bytes
Max core file size        0                    unlimited            bytes
Max resident set          unlimited            unlimited            bytes
Max processes             30442                30442                processes
Max open files            1024                 1048576              files
Max locked memory         65536                65536                bytes
Max address space         unlimited            unlimited            bytes
Max file locks            unlimited            unlimited            locks
Max pending signals       30442                30442                signals
Max msgqueue size         819200               819200               bytes
Max nice priority         0                    0
Max realtime priority     0                    0
Max realtime timeout      unlimited            unlimited            us
"""
    soft, hard = gc3libs.utils.parse_linux_proc_limits(data)
    # soft limits
    assert soft['max_realtime_timeout'] == None
    assert soft['max_realtime_priority'] == 0
    assert soft['max_nice_priority'] == 0
    assert soft['max_msgqueue_size'] == Memory(819200, unit=Memory.B)
    assert soft['max_pending_signals'] == 30442
    assert soft['max_file_locks'] == None
    assert soft['max_address_space'] == None
    assert soft['max_locked_memory'] == Memory(65536, unit=Memory.B)
    assert soft['max_open_files'] == 1024
    assert soft['max_processes'] == 30442
    assert soft['max_resident_set'] == None
    assert soft['max_core_file_size'] == Memory(0, unit=Memory.B)
    assert soft['max_stack_size'] == Memory(8388608, unit=Memory.B)
    assert soft['max_data_size'] == None
    assert soft['max_file_size'] == None
    assert soft['max_cpu_time'] == None
    # hard limits
    assert hard['max_realtime_timeout'] == None
    assert hard['max_realtime_priority'] == 0
    assert hard['max_nice_priority'] == 0
    assert hard['max_msgqueue_size'] == Memory(819200, unit=Memory.B)
    assert hard['max_pending_signals'] == 30442
    assert hard['max_file_locks'] == None
    assert hard['max_address_space'] == None
    assert hard['max_locked_memory'] == Memory(65536, unit=Memory.B)
    assert hard['max_open_files'] == 1048576
    assert hard['max_processes'] == 30442
    assert hard['max_resident_set'] == None
    assert hard['max_core_file_size'] == None
    assert hard['max_stack_size'] == None
    assert hard['max_data_size'] == None
    assert hard['max_file_size'] == None
    assert hard['max_cpu_time'] == None


class TestYieldAtNext(object):

    def test_YieldAtNext_yield(self):
        def generator_yield():
            yield 0
        g = gc3libs.utils.YieldAtNext(generator_yield())
        assert next(g) == 0

    def test_YieldAtNext_send(self):
        def generator_yield_send():
            val = (yield 1)
            yield val
        g = gc3libs.utils.YieldAtNext(generator_yield_send())
        assert next(g) == 1
        result = g.send('a sent value')
        assert result == None
        assert next(g) == 'a sent value'

    def test_YieldAtNext_send_iter(self):
        def generator_yield_send():
            val = (yield 0)
            while True:
                val = (yield val)
        g = gc3libs.utils.YieldAtNext(generator_yield_send())
        expected = list(range(0, 9))
        n = 0
        print ("expecting %d messages" % (len(expected),))
        for msg in g:
            print ("received: %s" % (msg,))
            # check msg
            assert msg == expected[n]
            # send another msg
            n += 1
            if n < len(expected):
                result = g.send(expected[n])
                assert result == None
            else:
                assert n == len(expected)
                print ("%d messages received, no more messages to send" % (n,))
                break

    def test_YieldAtNext_send_many(self):
        def generator_yield_send():
            val = (yield 0)
            while True:
                val = (yield val)
        g = gc3libs.utils.YieldAtNext(generator_yield_send())
        expected = list(range(1, 10))
        # consume one value to init the generator
        next(g)
        # send all messages
        for msg in expected:
            g.send(msg)
            print ("sent message '%s'" % (msg,))
        # receive them all
        print ("expecting %d messages back" % (len(expected),))
        for msg, expected_msg in zip(g, expected):
            print ("received: %s" % (msg,))
            # check msg
            assert msg == expected_msg

    def test_YieldAtNext_throw(self):
        def generator_yield_throw():
            try:
                # XXX: why do we need val = yield syntax here?
                # val = (yield 2)
                yield 2
            except RuntimeError:
                yield 'exception caught'
        g = gc3libs.utils.YieldAtNext(generator_yield_throw())
        assert next(g) == 2
        result = g.throw(RuntimeError)
        assert result == None
        assert next(g) == 'exception caught'

    def test_YieldAtNext_StopIteration(self):
        def generator_yield():
            yield 3
        g = gc3libs.utils.YieldAtNext(generator_yield())
        assert next(g) == 3
        with pytest.raises(StopIteration):
            next(g)


# main: run tests

if "__main__" == __name__:
    pytest.main(["-v", __file__])



#! /usr/bin/env python
"""
Specialized support for popular scientific applications.

Each application-specific class is packaged in a separate
submodule, which see for details.
"""
# Copyright (C) 2009-2012  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
__docformat__ = 'reStructuredText'


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="__init__",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   gtrac.py -- Front-end script for running the docking program rDock
#   over a list of ligand files.
#
#   Copyright (C) 2014, 2015  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gtrac.py --help`` for program usage
instructions.

Input parameters consists of:

...

Options:
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2015-02-17:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gtrac
    gtrac.GtracScript().run()

import os
import sys
import time
import tempfile
import re

import shutil
# import csv

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_CORES = 1
DEFAULT_MEMORY = Memory(3000,MB)

DEFAULT_REMOTE_INPUT_FOLDER="./"
DEFAULT_REMOTE_OUTPUT_FOLDER="./output"
DMRIC_PATTERN = "dmrirc"

## custom application class
class GtracApplication(Application):
    """
    """
    application_name = 'gtraci'

    def __init__(self, subject, subject_folder, dmrirc, **extra_args):

        self.output_dir = extra_args['output_dir']

        inputs = dict()
        outputs = dict()

        inputs[subject_folder] = DEFAULT_REMOTE_INPUT_FOLDER

        arguments = "trac-all -prep -c %s -debug" % dmrirc

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = [DEFAULT_REMOTE_OUTPUT_FOLDER],
            stdout = 'gtrac.log',
            join=True,
            **extra_args)

class GtracScript(SessionBasedScript):
    """

    The ``gtrac`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gtrac`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GtracApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GtracApplication,
            )

    def setup_args(self):

        self.add_param('input_data', type=str,
                       help="Root localtion of input data. "
                       "Note: expected folder structure: "
                       " 1 subfodler for each subject. "
                       " In each subject folder, "
                       " 1 subfolder for each TimePoint. "
                       " Each TimePoint folder should contain 2 input "
                       "NFTI files.")

    def new_tasks(self, extra):
        """
        For each input folder, create an instance of GtracApplication
        """
        tasks = []

        for (subject_folder,subject_name, dmric) in self.get_input_subject_folder(self.params.input_data):

            # extract root folder name to be used as jobname
            extra_args = extra.copy()
            extra_args['jobname'] = subject_name

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                        'run_%s' % subject_name)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                        'run_%s' % subject_name)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                        'run_%s' % subject_name)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                        'run_%s' % subject_name)

            tasks.append(GtracApplication(
                subject_name,
                os.path.join(self.params.input_data,subject_folder),
                dmric,
                **extra_args))

        return tasks

    def get_input_subject_folder(self, input_folder):
        """
        Check and validate input subfolders
        XXX: for the time being just pass
        """

        for r,d,f in os.walk(input_folder):
            for infile in f:
                if infile.startswith(DMRIC_PATTERN):
                    yield (os.path.abspath(r),r,infile)

        # return os.listdir(input_folder)
